# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a way to interact with Speos feature: ground plane."""

from __future__ import annotations

from difflib import SequenceMatcher
from typing import List, Optional

from ansys.speos.core.kernel.scene import ProtoScene
import ansys.speos.core.project as project
import ansys.speos.core.proto_message_utils as proto_message_utils


class GroundPlane:
    """Speos feature: ground plane.

    Only usable when there is at least one Ambient Environment Source in the project.

    Parameters
    ----------
    project : project.Project
        Project that will own the feature.
    """

    def __init__(self, project: project.Project, ground: Optional[ProtoScene.GroundPlane] = None):
        self._project = project
        self._name = ""

        if ground is None:
            # Create local ground
            self._ground = ProtoScene.GroundPlane()
            self.ground_origin = [0.0, 0.0, 0.0]
            self.ground_zenith = [0.0, 0.0, 1.0]
            self.ground_height = 1000.0
            self._committed = False
        else:
            # Retrieve ground from input
            self._committed = True
            self._ground = ground

    @property
    def ground_origin(self) -> List[float]:
        """Ground origin.

        This property gets and sets the origin of the ground plane.
        Default as [0, 0, 0]

        Parameters
        ----------
        value: List[float]
            Ground origin.

        Returns
        -------
        List[float]
            Ground origin.
        """
        return self._ground.ground_origin

    @ground_origin.setter
    def ground_origin(self, value: List[float]) -> None:
        self._ground.ground_origin[:] = value

    @property
    def ground_zenith(self) -> List[float]:
        """Zenith direction.

        This property gets and sets the zenith direction of the ground plane.
        Default as [0, 0, 1]

        Parameters
        ----------
        value: List[float]
            Zenith direction.

        Returns
        -------
        List[float]
            Zenith direction.
        """
        return self._ground.zenith_direction

    @ground_zenith.setter
    def ground_zenith(self, value: List[float]) -> None:
        self._ground.zenith_direction[:] = value

    @property
    def ground_height(self) -> float:
        """Ground height.

        This property gets and sets the height of the ground plane.
        Default as 1000.0

        Parameters
        ----------
        value: float
            Ground height.

        Returns
        -------
        float
            Ground height.
        """
        return self._ground.ground_height

    @ground_height.setter
    def ground_height(self, value: float) -> None:
        self._ground.ground_height = value

    def _to_dict(self) -> dict:
        out_dict = {}

        if self._project.scene_link:
            scene_data = self._project.scene_link.get()

            if self._committed:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._project.client, message=scene_data.ground
                )
            else:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._project.client, message=self._ground
                )
        else:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._project.client, message=self._ground
            )

        proto_message_utils._replace_properties(json_dict=out_dict)

        return out_dict

    def get(self, key: str = "") -> str | dict:
        """Get dictionary corresponding to the ground plane - read only.

        Parameters
        ----------
        key: str

        Returns
        -------
        str | dict
        """
        if key == "":
            return self._to_dict()
        info = proto_message_utils._value_finder_key_startswith(dict_var=self._to_dict(), key=key)
        content = list(info)
        if len(content) != 0:
            content.sort(
                key=lambda x: SequenceMatcher(None, x[0], key).ratio(),
                reverse=True,
            )
            return content[0][1]
        info = proto_message_utils._flatten_dict(dict_var=self._to_dict())
        print("Used key: {} not found in key list: {}.".format(key, info.keys()))

    def __str__(self):
        """Return the string representation of the ground plane property."""
        out_str = ""

        if self._project.scene_link:
            if not self._committed:
                out_str += "local: "
        else:
            out_str += "local: "

        out_str += proto_message_utils.dict_to_str(dict=self._to_dict())
        return out_str

    def commit(self) -> GroundPlane:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.ground_plane.GroundPlane
            Ground plane feature.
        """
        # This boolean allows to keep track if the ground plane is committed
        self._committed = True

        # Update the scene with the ground plane
        if self._project.scene_link:
            scene_data = self._project.scene_link.get()  # retrieve scene data

            # In case the ground is different from what is stored on server -> update
            if scene_data.ground != self._ground:
                scene_data.ground.CopyFrom(self._ground)
                self._project.scene_link.set(data=scene_data)  # update scene data

        return self

    def reset(self) -> GroundPlane:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.ground_plane.GroundPlane
            Ground plane feature.
        """
        # Reset ground plane
        if self._project.scene_link is not None:
            scene_data = self._project.scene_link.get()  # retrieve scene data
            self._ground = scene_data.ground  # store locally the ground from server data

        return self

    def delete(self) -> GroundPlane:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.ground_plane.GroundPlane
            Ground plane feature.
        """
        # Remove the ground from the scene
        scene_data = self._project.scene_link.get()  # retrieve scene data

        if scene_data.HasField("ground"):
            scene_data.ClearField("ground")
            self._project.scene_link.set(data=scene_data)  # update scene data

        # _committed to false -> this feature is no longer committed
        self._committed = False

        return self



# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a wrapped abstraction of the gRPC proto API definition and stubs."""

from typing import Iterator, List

from ansys.api.speos.part.v1 import (
    face_pb2 as messages,
    face_pb2_grpc as service,
)
from grpc import RpcError

from ansys.speos.core.generic.general_methods import min_speos_version
from ansys.speos.core.kernel.crud import CrudItem, CrudStub
from ansys.speos.core.kernel.proto_message_utils import protobuf_message_to_str

ProtoFace = messages.Face
"""Face protobuf class : ansys.api.speos.part.v1.face_pb2.Face"""
ProtoFace.__str__ = lambda self: protobuf_message_to_str(self)


class FaceLink(CrudItem):
    """Link object for job in database.

    Parameters
    ----------
    db : ansys.speos.core.kernel.face.FaceStub
        Database to link to.
    key : str
        Key of the face in the database.
    """

    def __init__(self, db, key: str):
        super().__init__(db, key)

    def __str__(self) -> str:
        """Return the string representation of the face."""
        return str(self.get())

    def get(self) -> ProtoFace:
        """Get the datamodel from database.

        Returns
        -------
        face.Face
            Face datamodel.
        """
        return self._stub.read(self)

    def set(self, data: ProtoFace) -> None:
        """Change datamodel in database.

        Parameters
        ----------
        data : face.Face
            New Face datamodel.
        """
        self._stub.update(self, data)

    def delete(self) -> None:
        """Remove datamodel from database."""
        self._stub.delete(self)


class FaceStub(CrudStub):
    """
    Database interactions for face.

    Parameters
    ----------
    channel : grpc.Channel
        Channel to use for the stub.

    Examples
    --------
    The best way to get a FaceStub is to retrieve it from SpeosClient via faces() method.
    Like in the following example:

    >>> from ansys.speos.core.speos import Speos
    >>> speos = Speos()
    >>> face_db = speos.client.faces()

    """

    def __init__(self, channel):
        super().__init__(stub=service.FacesManagerStub(channel=channel))
        self._actions_stub = service.FaceActionsStub(channel=channel)
        self._is_batch_available = self._check_if_batch_available()

    def _check_if_batch_available(self) -> bool:
        try:
            for reserve_faces_res in self._actions_stub.ReserveFaces(
                FaceStub._reserve_face_iterator([ProtoFace(name="tmp")])
            ):
                FaceLink(self, reserve_faces_res.guids[0]).delete()
            return True
        except RpcError:
            return False

    @min_speos_version(25, 2, 0)
    def create_batch(self, message_list: List[ProtoFace]) -> List[FaceLink]:
        """Create new entries.

        Parameters
        ----------
        message_list : List[face.Face]
            List of datamodels for the new entries.

        Returns
        -------
        List[ansys.speos.core.kernel.face.FaceLink]
            List pf link objects created.
        """
        if not self._is_batch_available:
            raise NotImplementedError("Please use a Speos Version of 2025 R2 SP0 or higher.")

        reserve_faces_res = self._actions_stub.ReserveFaces(
            FaceStub._reserve_face_iterator(message_list)
        )

        guids = []
        for res in reserve_faces_res:
            for guid in res.guids:
                guids.append(guid)

        chunk_iterator = FaceStub._faces_to_chunks(
            guids=guids, message_list=message_list, nb_items=128 * 1024
        )
        self._actions_stub.Upload(chunk_iterator)

        return [FaceLink(self, guid) for guid in guids]

    def create(self, message: ProtoFace) -> FaceLink:
        """Create a new entry.

        Parameters
        ----------
        message : face.Face
            Datamodel for the new entry.

        Returns
        -------
        ansys.speos.core.kernel.face.FaceLink
            Link object created.
        """
        resp = CrudStub.create(self, messages.Create_Request(face=ProtoFace(name="tmp")))

        chunk_iterator = FaceStub._faces_to_chunks(
            guids=[resp.guid], message_list=[message], nb_items=128 * 1024
        )
        self._actions_stub.Upload(chunk_iterator)
        return FaceLink(self, resp.guid)

    @min_speos_version(25, 2, 0)
    def read_batch(self, refs: List[FaceLink]) -> List[ProtoFace]:
        """Get existing entries.

        Parameters
        ----------
        refs : List[ansys.speos.core.kernel.face.FaceLink]
            List of link objects to read.

        Returns
        -------
        List[face.Face]
            Datamodels of the entries.
        """
        if not self._is_batch_available:
            raise NotImplementedError("Please use a Speos Version of 2025 R2 SP0 or higher.")

        for ref in refs:
            if not ref.stub == self:
                raise ValueError("FaceLink is not on current database. Key=" + ref.key)
        chunks = self._actions_stub.Download(
            request=messages.Download_Request(guids=[ref.key for ref in refs])
        )
        return FaceStub._chunks_to_faces(chunks)

    def read(self, ref: FaceLink) -> ProtoFace:
        """Get an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.face.FaceLink
            Link object to read.

        Returns
        -------
        face.Face
            Datamodel of the entry.
        """
        if not ref.stub == self:
            raise ValueError("FaceLink is not on current database. Key=" + ref.key)

        chunks = self._actions_stub.Download(request=messages.Download_Request(guid=ref.key))
        return FaceStub._chunks_to_faces(chunks)[0]

    @min_speos_version(25, 2, 0)
    def update_batch(self, refs: List[FaceLink], data: List[ProtoFace]) -> None:
        """Change existing entries.

        Parameters
        ----------
        ref : List[ansys.speos.core.kernel.face.FaceLink]
            Link objects to update.

        data : List[face.Face]
            New datamodels for the entries.
        """
        if not self._is_batch_available:
            raise NotImplementedError("Please use a Speos Version of 2025 R2 SP0 or higher.")

        for ref in refs:
            if not ref.stub == self:
                raise ValueError("FaceLink is not on current database")

        chunk_iterator = FaceStub._faces_to_chunks(
            guids=[ref.key for ref in refs], message_list=data, nb_items=128 * 1024
        )
        self._actions_stub.Upload(chunk_iterator)

    def update(self, ref: FaceLink, data: ProtoFace) -> None:
        """Change an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.face.FaceLink
            Link object to update.

        data : face.Face
            New datamodel for the entry.
        """
        if not ref.stub == self:
            raise ValueError("FaceLink is not on current database")

        chunk_iterator = FaceStub._faces_to_chunks(
            guids=[ref.key], message_list=[data], nb_items=128 * 1024
        )
        self._actions_stub.Upload(chunk_iterator)

    def delete(self, ref: FaceLink) -> None:
        """Remove an existing entry.

        Parameters
        ----------
        ref : ansys.speos.core.kernel.face.FaceLink
            Link object to delete.
        """
        if not ref.stub == self:
            raise ValueError("FaceLink is not on current database")
        CrudStub.delete(self, messages.Delete_Request(guid=ref.key))

    def list(self) -> List[FaceLink]:
        """List existing entries.

        Returns
        -------
        List[ansys.speos.core.kernel.face.FaceLink]
            Link objects.
        """
        guids = CrudStub.list(self, messages.List_Request()).guids
        return list(map(lambda x: FaceLink(self, x), guids))

    @staticmethod
    def _reserve_face_iterator(
        message_list: List[ProtoFace],
    ) -> Iterator[messages.ReserveFace_Request]:
        for message in message_list:
            yield messages.ReserveFace_Request(faces=[ProtoFace(name="tmp")])

    @staticmethod
    def _faces_to_chunks(
        guids: List[str], message_list: List[ProtoFace], nb_items: int
    ) -> Iterator[messages.Chunk]:
        for guid, message in zip(guids, message_list):
            for j in range(4):
                if j == 0:
                    chunk_face_header = messages.Chunk(
                        face_header=messages.Chunk.FaceHeader(
                            guid=guid,
                            name=message.name,
                            description=message.description,
                            metadata=message.metadata,
                            sizes=[
                                len(message.vertices),
                                len(message.facets),
                                len(message.vertices_data),
                            ],
                        )
                    )
                    yield chunk_face_header
                elif j == 1:
                    for i in range(0, len(message.vertices), nb_items):
                        chunk_vertices = messages.Chunk(
                            vertices=messages.Chunk.Vertices(
                                data=message.vertices[i : i + nb_items]
                            )
                        )
                        yield chunk_vertices
                elif j == 2:
                    for i in range(0, len(message.facets), nb_items):
                        chunk_facets = messages.Chunk(
                            facets=messages.Chunk.Facets(data=message.facets[i : i + nb_items])
                        )
                        yield chunk_facets
                elif j == 3:
                    for i in range(0, len(message.normals), nb_items):
                        chunk_normals = messages.Chunk(
                            normals=messages.Chunk.Normals(data=message.normals[i : i + nb_items])
                        )
                        yield chunk_normals

    @staticmethod
    def _chunks_to_faces(chunks: messages.Chunk) -> List[ProtoFace]:
        out_faces = []
        out_face = ProtoFace()
        for chunk in chunks:
            if chunk.HasField("face_header"):
                if out_face != ProtoFace():  # Add face each time a new one starts
                    out_faces.append(out_face)
                    out_face = ProtoFace()
                out_face.name = chunk.face_header.name
                out_face.description = chunk.face_header.description
                out_face.metadata.update(chunk.face_header.metadata)
            if chunk.HasField("vertices"):
                out_face.vertices.extend(chunk.vertices.data)
            if chunk.HasField("facets"):
                out_face.facets.extend(chunk.facets.data)
            if chunk.HasField("normals"):
                out_face.normals.extend(chunk.normals.data)

        out_faces.append(out_face)  # Don't forget to add last face
        return out_faces



# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""The lxp module contains classes and functions to simplify the interaction with ray data.

Ray data is provided as lpf file.
LPF files contain a set of simulated rays with all their intersections and properties.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

import ansys.api.speos.lpf.v2.lpf_file_reader_pb2 as lpf_file_reader__v2__pb2
import ansys.api.speos.lpf.v2.lpf_file_reader_pb2_grpc as lpf_file_reader__v2__pb2_grpc

from ansys.speos.core.generic.general_methods import graphics_required, wavelength_to_rgb
from ansys.speos.core.project import Project, Speos

if TYPE_CHECKING:  # pragma: no cover
    from ansys.tools.visualization_interface import Plotter
try:
    from ansys.speos.core.generic.general_methods import run_if_graphics_required

    run_if_graphics_required(warning=True)
except ImportError as err:  # pragma: no cover
    raise err

ERROR_IDS = [7, 8, 9, 10, 11, 12, 13, 14, 15]
"""Intersection types indicating an error state."""

NO_ERROR_IDS = [0, 1, 2, 3, 4, 5, 6, 16, -7, -6, -5, -5, -4, -3, -2, -1]
"""Intersection types indicating a correct ray state."""


class RayPath:
    """Framework representing a singular ray path.

    Parameters
    ----------
    raypath : ansys.api.speos.lpf.v2.lpf_file_reader__v2__pb2.RayPath
        RayPath object
    sensor_contribution : bool
        Defines if sensor contributions are stored within the data.
        By default ``False``.
    """

    def __init__(
        self,
        raypath: lpf_file_reader__v2__pb2.RayPath,
        sensor_contribution: bool = False,
    ):
        self._nb_impacts = len(raypath.impacts)
        self._impacts = [[inter.x, inter.y, inter.z] for inter in raypath.impacts]
        self._wl = raypath.wavelengths[0]
        self._body_ids = raypath.body_context_ids
        self._face_ids = raypath.unique_face_ids
        self._last_direction = [
            raypath.lastDirection.x,
            raypath.lastDirection.y,
            raypath.lastDirection.z,
        ]
        self._intersection_type = raypath.interaction_statuses
        if sensor_contribution:
            self._sensor_contribution = [
                {
                    "sensor_id": sc.sensor_id,
                    "position": [sc.coordinates.x, sc.coordinates.y],
                }
                for sc in raypath.sensor_contributions
            ]
        else:
            self._sensor_contribution = None

    @property
    def nb_impacts(self) -> int:
        """Number of impacts contained in ray path.

        Returns
        -------
        int
            Number of impacts
        """
        return self._nb_impacts

    @property
    def impacts(self) -> list[list[float]]:
        """XYZ coordinates for each impact.

        Returns
        -------
        list[list[float]]
            list containing the impact coordinates [[x0,y0,z0],[x1,y1,z1],...]
        """
        return self._impacts

    @property
    def wl(self) -> float:
        """Wavelength of the ray.

        Returns
        -------
        float
            Wavelength in nm
        """
        return self._wl

    @property
    def body_ids(self) -> list[int]:
        """Body ID of interacted body for each impact.

        Returns
        -------
        list[int]
            List of body IDs for each impact.
        """
        return self._body_ids

    @property
    def face_ids(self) -> list[int]:
        """Face ID of interacted body for each impact.

        Returns
        -------
        list[int]
            List of face IDs for each impact.
        """
        return self._face_ids

    @property
    def last_direction(self) -> list[float]:
        """Last direction of the ray.

        Returns
        -------
        list[float]
            Last direction of the rays as list[x,y,z].
        """
        return self._last_direction

    @property
    def intersection_type(self) -> list[int]:
        """Intersection type of the ray for each impact.

        Returns
        -------
        list[int]
            Intersection type at each impact.

        Notes
        -----
        Available intersection types:

        - StatusAbsorbed = 0
        - StatusSpecularTransmitted = 1
        - StatusGaussianTransmitted = 2
        - StatusLambertianTransmitted = 3
        - StatusVolumicDiffused = 4
        - StatusJustEmitted = 5
        - StatusDiracTransmitted = 6
        - StatusError = 7
        - StatusErrorVolumicBodyNotClosed = 8
        - StatusErrorVolumeConflict = 9
        - StatusError2DTangency = 10
        - StatusError2DIntersect3DWarning = 11
        - StatusErrorNonOpticalMaterial = 12
        - StatusErrorIntersection = 13
        - StatusErrorNonOpticalMaterialAtEmission = 14
        - StatusError3DTextureSupportTangency = 15
        - StatusLast = 16
        - StatusFirst = -7
        - StatusDiracReflected = -6
        - StatusReserved = -5
        - StatusGrinStep = -4
        - StatusLambertianReflected = -3
        - StatusGaussianReflected = -2
        - StatusSpecularReflected = -1
        """
        return self._intersection_type

    @property
    def sensor_contribution(self) -> Union[None, list[dict]]:
        """Provide the sensor contribution information for each sensor.

        Returns
        -------
        Union[None, list[dict]]
            If no sensor contribution, None will be returned. If there is sensor contribution, \
            a dictionary with the following information is returned:\
            {“sensor_id”: sc.sensor_id,
            “position”: [sc.coordinates.x, sc.coordinates.y]}
        """
        return self._sensor_contribution

    def get(self, key=""):
        """Retrieve any information from the RayPath object.

        Parameters
        ----------
        key : str
            Name of the property.

        Returns
        -------
        property
            Values/content of the associated property.
        """
        data = {k: v.fget(self) for k, v in RayPath.__dict__.items() if isinstance(v, property)}
        if key == "":
            return data
        elif data.get(key):
            return data.get(key)
        else:
            print("Used key: {} not found in key list: {}.".format(key, data.keys()))

    def __str__(self):
        """Create string representation of a RayPath."""
        return str(self.get())


class LightPathFinder:
    """Define an interface to read LPF files.

    LPF files contain a set of simulated rays including their intersections and properties.

    Parameters
    ----------
    speos : ansys.speos.core.speos.Speos
        Speos Session (connected to Speos gRPC server).
    path : str
        Path to the LPF file to be opened.

    """

    def __init__(self, speos: Speos, path: str):
        self.client = speos.client
        """Speos instance client"""
        self._stub = lpf_file_reader__v2__pb2_grpc.LpfFileReader_MonoStub(self.client.channel)
        self.__open(path)
        self._data = self._stub.GetInformation(
            lpf_file_reader__v2__pb2.GetInformation_Request_Mono()
        )
        self._nb_traces = self._data.nb_of_traces
        self._nb_xmps = self._data.nb_of_xmps
        self._has_sensor_contributions = self._data.has_sensor_contributions
        self._sensor_names = self._data.sensor_names
        self._rays = self.__parse_traces()
        self._filtered_rays = []

    @property
    def nb_traces(self) -> int:
        """Number of light paths within LPF data set."""
        return self._nb_traces

    @property
    def nb_xmps(self) -> int:
        """Number of sensors involved within LPF data set."""
        return self._nb_xmps

    @property
    def has_sensor_contributions(self) -> bool:
        """Define if a LPF file contains information regarding the sensor contribution."""
        return self._has_sensor_contributions

    @property
    def sensor_names(self) -> list[str]:
        """List of involved sensor names."""
        return self._sensor_names

    @property
    def rays(self) -> list[RayPath]:
        """List ray paths within LPF file."""
        return self._rays

    @property
    def filtered_rays(self) -> list[RayPath]:
        """List of filtered ray paths."""
        return self._filtered_rays

    def __str__(self):
        """Create string representation of LightPathFinder."""
        return str(
            {
                k: v.fget(self)
                for k, v in LightPathFinder.__dict__.items()
                if isinstance(v, property) and "rays" not in k
            }
        )

    def __open(self, path: str):
        """Open LPF file.

        Parameters
        ----------
        path : str
            Path to file
        """
        self._stub.InitLpfFileName(
            lpf_file_reader__v2__pb2.InitLpfFileName_Request_Mono(lpf_file_uri=path)
        )

    def __parse_traces(self) -> list[RayPath]:
        """Read all ray paths from lpf dataset.

        Returns
        -------
            list[script.RayPath]

        """
        raypaths = []
        for rp in self._stub.Read(lpf_file_reader__v2__pb2.Read_Request_Mono()):
            raypaths.append(RayPath(rp, self._has_sensor_contributions))
        return raypaths

    def __filter_by_last_intersection_types(self, options: list[int], new=True):
        """Filter ray paths based on last intersection types.

        Populate filtered_rays property.
        """
        if new:
            self._filtered_rays = []
            for ray in self._rays:
                if int(ray.intersection_type[-1]) in options:
                    self._filtered_rays.append(ray)
        else:
            temp_rays = self._filtered_rays
            self._filtered_rays = []
            for ray in temp_rays:
                if int(ray.intersection_type[-1]) in options:
                    self._filtered_rays.append(ray)

    def filter_by_face_ids(self, options: list[int], new=True) -> LightPathFinder:
        """Filter ray paths based on face IDs and populates filtered_rays property.

        Parameters
        ----------
        options : list[int]
            List of face IDs.
        new : bool
            Define if a new filter is created or an existing filter is filtered.

        Returns
        -------
        ansys.speos.core.lxp.LightPathFinder
            LightPathFinder Instance.
        """
        if new:
            self._filtered_rays = []
            for ray in self._rays:
                if any([faceid in options for faceid in ray.face_ids]):
                    self._filtered_rays.append(ray)
        else:
            temp_rays = self._filtered_rays
            self._filtered_rays = []
            for ray in temp_rays:
                if any([faceid in options for faceid in ray.face_ids]):
                    self._filtered_rays.append(ray)
        return self

    def filter_by_body_ids(self, options: list[int], new=True) -> LightPathFinder:
        """Filter ray paths based on body IDs and populates filtered_rays property.

        Parameters
        ----------
        options : list[int]
            List of body IDs.
        new : bool
            Define if a new filter is created or an existing filter is filtered.

        Returns
        -------
        ansys.speos.core.lxp.LightPathFinder
            LightPathFinder Instance.
        """
        if new:
            self._filtered_rays = []
            for ray in self._rays:
                if any([body_id in options for body_id in ray.body_ids]):
                    self._filtered_rays.append(ray)
        else:
            temp_rays = self._filtered_rays
            self._filtered_rays = []
            for ray in temp_rays:
                if any([body_id in options for body_id in ray.body_ids]):
                    self._filtered_rays.append(ray)
        return self

    def filter_error_rays(self) -> LightPathFinder:
        """Filter ray paths and only shows rays in error.

        Returns
        -------
        ansys.speos.core.lxp.LightPathFinder
            LightPathFinder Instance.
        """
        self.__filter_by_last_intersection_types(options=ERROR_IDS)
        return self

    def remove_error_rays(self) -> LightPathFinder:
        """Filter rays and only shows rays not in error.

        Returns
        -------
        ansys.speos.core.lxp.LightPathFinder
            LightPathFinder Instance.
        """
        self.__filter_by_last_intersection_types(options=NO_ERROR_IDS)
        return self

    @staticmethod
    @graphics_required
    def __add_ray_to_pv(plotter: Plotter, ray: RayPath, max_ray_length: float):
        """Add a ray to pyvista plotter.

        Parameters
        ----------
        plotter : Plotter
            Ansys plotter object to which rays should be added.
        ray : script.RayPath
            RayPath object which contains ray information to be added.
        max_ray_length : float
            Length of the last ray.
        """
        import pyvista as pv

        temp = ray.impacts.copy()
        if not 7 <= ray.intersection_type[-1] <= 15:
            temp.append(
                [
                    ray.impacts[-1][0] + max_ray_length * ray.last_direction[0],
                    ray.impacts[-1][1] + max_ray_length * ray.last_direction[1],
                    ray.impacts[-1][2] + max_ray_length * ray.last_direction[2],
                ]
            )
        if len(ray.impacts) > 2:
            mesh = pv.MultipleLines(temp)
        else:
            mesh = pv.Line(temp[0], temp[1])
        plotter.plot(mesh, color=wavelength_to_rgb(ray.wl), line_width=2)

    @graphics_required
    def preview(
        self,
        nb_ray: int = 100,
        max_ray_length: float = 50.0,
        ray_filter: bool = False,
        project: Project = None,
        screenshot: Optional[Union[str, Path]] = None,
    ) -> LightPathFinder:
        """Preview LPF file with pyvista.

        Parameters
        ----------
        nb_ray : int
            Number of rays to be visualized.
        max_ray_length : float
            Length of last ray.
        ray_filter : bool
            Boolean to decide if filtered rays or all rays should be shown.
        project : ansys.speos.core.project.Project
            Speos Project/Geometry to be added to pyvista visualisation.
        screenshot : str or Path or ``None``
            Path to save a screenshot of the plotter. If defined Plotter will only create the
            screenshot

        Returns
        -------
        ansys.speos.core.lxp.LightPathFinder
            LightPathFinder Instance.

        Notes
        -----
        Please use the ``q``-key to close the plotter as some
        operating systems (namely Windows) will experience issues
        saving a screenshot if the exit button in the GUI is pressed.
        """
        from ansys.tools.visualization_interface import Plotter

        if ray_filter:
            if len(self._filtered_rays) > 0:
                temp_rays = self._filtered_rays
            else:
                print("no filtered rays")
                temp_rays = self._rays
        else:
            temp_rays = self._rays
        if not project:
            plotter = Plotter()
            if nb_ray > len(temp_rays):
                for ray in temp_rays:
                    self.__add_ray_to_pv(plotter, ray, max_ray_length)
            else:
                for i in range(nb_ray):
                    self.__add_ray_to_pv(plotter, temp_rays[i], max_ray_length)
        else:
            plotter = project._create_preview(viz_args={"opacity": 0.5})
            if nb_ray > len(temp_rays):
                for ray in temp_rays:
                    self.__add_ray_to_pv(plotter, ray, max_ray_length)
            else:
                for i in range(nb_ray):
                    self.__add_ray_to_pv(plotter, temp_rays[i], max_ray_length)
        plotter.show(screenshot=screenshot)
        return self



# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""General methods and helpers collection.

this includes decorator and methods
"""

from __future__ import annotations

from collections.abc import Collection
from functools import lru_cache, wraps
import os
from pathlib import Path
from typing import List, Optional, Union, cast
import warnings

from ansys.tools.common.path import get_available_ansys_installations
import numpy as np

from ansys.speos.core.generic.constants import DEFAULT_VERSION

_GRAPHICS_AVAILABLE = None

GRAPHICS_ERROR = (
    "Preview unsupported without 'ansys-tools-visualization_interface' installed. "
    "You can install this using `pip install ansys-speos-core[graphics]`."
)

VERSION_ERROR = "The pySpeos feature : {feature_name} needs a Speos Version of {version} or higher."


def deprecate_kwargs(old_arguments: dict, removed_version="0.3.0"):
    """Issues deprecation warnings for arguments.

    Parameters
    ----------
    old_arguments : dict
        key old argument value new argument name
    removed_version : str
        Release version with which argument support will be removed
        By Default, next major release

    """

    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            func_name = function.__name__
            for alias, new in old_arguments.items():
                if alias in kwargs:
                    if new in kwargs:
                        msg = f"{func_name} received both {alias} and {new} as arguments!\n"
                        msg += f"{alias} is deprecated, use {new} instead."
                        raise TypeError(msg)
                    msg = f"Argument `{alias}` is deprecated for method `{func_name}`; it will be "
                    msg += f"removed with Release v{removed_version}. Please use `{new}` instead."
                    kwargs[new] = kwargs.pop(alias)
                    warnings.warn(msg, DeprecationWarning, stacklevel=2)
            retval = function(*args, **kwargs)
            return retval

        return wrapper

    return decorator


@lru_cache
def run_if_graphics_required(warning=False):
    """Check if graphics are available."""
    global _GRAPHICS_AVAILABLE
    if _GRAPHICS_AVAILABLE is None:
        try:
            from ansys.tools.visualization_interface import Plotter  # noqa: F401
            import pyvista as pv  # noqa: F401

            _GRAPHICS_AVAILABLE = True
        except ImportError:  # pragma: no cover
            _GRAPHICS_AVAILABLE = False

    if _GRAPHICS_AVAILABLE is False and warning is False:  # pragma: no cover
        raise ImportError(GRAPHICS_ERROR)
    elif _GRAPHICS_AVAILABLE is False:  # pragma: no cover
        warnings.warn(GRAPHICS_ERROR)


def graphics_required(method):
    """Decorate a method as requiring graphics.

    Parameters
    ----------
    method : callable
        Method to decorate.

    Returns
    -------
    callable
        Decorated method.
    """

    def wrapper(*args, **kwargs):
        run_if_graphics_required()
        return method(*args, **kwargs)

    return wrapper


def magnitude_vector(vector: Collection[float]) -> float:
    """Compute the magnitude (length) of a 2D or 3D vector using NumPy.

    Parameters
    ----------
    vector: List[float]
        A 2D or 3D vector as a list [x, y] or [x, y, z].

    Returns
    -------
    float
        The magnitude (length) of the vector.
    """
    vector_np = np.array(vector, dtype=float)
    if vector_np.size not in (2, 3):
        raise ValueError("Input vector must be either 2D or 3D")
    return float(np.linalg.norm(vector_np))


def normalize_vector(vector: Collection[float]) -> List[float]:
    """
    Normalize a 2D or 3D vector to have a length of 1 using NumPy.

    Parameters
    ----------
    vector: List[float]
        A vector as a list [x, y] for 2D or [x, y, z] for 3D.

    Returns
    -------
    List[float]
        The normalized vector.
    """
    vector_np = np.array(vector, dtype=float)
    if vector_np.size not in (2, 3):
        raise ValueError("Input vector must be either 2D or 3D")

    magnitude = magnitude_vector(vector_np)
    if magnitude == 0:
        raise ValueError("Cannot normalize the zero vector")

    return cast(List[float], (vector_np / magnitude).tolist())


def error_no_install(install_path: Union[Path, str], version: Union[int, str]):
    """Raise error that installation was not found at a location.

    Parameters
    ----------
    install_path : Union[Path, str]
        Installation Path
    version : Union[int, str]
        Version
    """
    raise FileNotFoundError(
        f"Ansys Speos RPC server installation not found at {install_path}. "
        f"Please define AWP_ROOT{version} environment variable"
    )


def retrieve_speos_install_dir(
    speos_rpc_path: Optional[Union[Path, str]] = None, version: str = DEFAULT_VERSION
) -> Path:
    """Retrieve Speos install location based on Path or Environment.

    Parameters
    ----------
    speos_rpc_path : Optional[str, Path]
        location of Speos rpc executable
    version : Union[str, int]
        The Speos server version to run, in the 3 digits format, such as "242".
        If unspecified, the version will be chosen as
        ``ansys.speos.core.kernel.client.LATEST_VERSION``.

    """
    if not speos_rpc_path:
        speos_rpc_path = ""

    if not speos_rpc_path or not Path(speos_rpc_path).exists():
        if not Path(speos_rpc_path).exists():
            warnings.warn(
                "Provided executable location not found, looking for local installation",
                UserWarning,
            )
        versions = get_available_ansys_installations()
        ansys_loc = versions.get(int(version), "")
        if not ansys_loc:
            ansys_loc = os.environ.get("AWP_ROOT{}".format(version), "")
            if not ansys_loc:
                error_no_install(speos_rpc_path, int(version))

        speos_rpc_path = Path(ansys_loc) / "Optical Products" / "SPEOS_RPC"
    elif Path(speos_rpc_path).is_file():
        if "SpeosRPC_Server" not in Path(speos_rpc_path).name:
            error_no_install(speos_rpc_path, int(version))
        else:
            speos_rpc_path = Path(speos_rpc_path).parent

    speos_rpc_path = Path(speos_rpc_path)
    if os.name == "nt":
        speos_exec = speos_rpc_path / "SpeosRPC_Server.exe"
    else:
        speos_exec = speos_rpc_path / "SpeosRPC_Server.x"
    if not speos_exec.is_file():
        error_no_install(speos_rpc_path, int(version))
    return speos_rpc_path


def wavelength_to_rgb(wavelength: float, gamma: float = 0.8) -> [int, int, int, int]:
    """Convert a given wavelength of light to an approximate RGB color value.

    The wavelength must be given in nanometers in the range from 380 nm to 750 nm.
    Based on the code from http://www.physics.sfasu.edu/astro/color/spectra.html

    Parameters
    ----------
    wavelength : float
        Wavelength in nanometer between 380-750 nm
    gamma : float
        Gamma value.
        By default : ``0.8``
    """
    wavelength = float(wavelength)
    if 380 <= wavelength <= 440:
        attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
        r = ((-(wavelength - 440) / (440 - 380)) * attenuation) ** gamma
        g = 0.0
        b = (1.0 * attenuation) ** gamma
    elif 440 <= wavelength <= 490:
        r = 0.0
        g = ((wavelength - 440) / (490 - 440)) ** gamma
        b = 1.0
    elif 490 <= wavelength <= 510:
        r = 0.0
        g = 1.0
        b = (-(wavelength - 510) / (510 - 490)) ** gamma
    elif 510 <= wavelength <= 580:
        r = ((wavelength - 510) / (580 - 510)) ** gamma
        g = 1.0
        b = 0.0
    elif 580 <= wavelength <= 645:
        r = 1.0
        g = (-(wavelength - 645) / (645 - 580)) ** gamma
        b = 0.0
    elif 645 <= wavelength <= 750:
        attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
        r = (1.0 * attenuation) ** gamma
        g = 0.0
        b = 0.0
    else:
        r = 0.0
        g = 0.0
        b = 0.0
    r *= 255
    g *= 255
    b *= 255
    return [int(r), int(g), int(b), 255]


def min_speos_version(major: int, minor: int, service_pack: int):
    """Raise version warning.

    Parameters
    ----------
    major : int
        Major release version, e.g. 25
    minor : int
        Minor release version e.g. 1
    service_pack : int
        Service Pack version e.g. 3
    """
    version = f"20{major} R{minor} SP{service_pack}"

    def decorator(function):
        def wrapper(*args, **kwargs):
            if function.__qualname__.endswith("__init__"):
                name = function.__qualname__[:-9]
            else:
                name = function.__qualname__
            warnings.warn(VERSION_ERROR.format(version=version, feature_name=name), stacklevel=2)
            return function(*args, **kwargs)

        return wrapper

    return decorator



# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a way to interact with Speos feature: Source."""

from __future__ import annotations

import datetime
from difflib import SequenceMatcher
from pathlib import Path
from typing import List, Mapping, Optional, Union
import uuid

from ansys.api.speos.scene.v2 import scene_pb2
from ansys.api.speos.source.v1 import source_pb2
import numpy as np

from ansys.speos.core import (
    project as project,
    proto_message_utils as proto_message_utils,
)
import ansys.speos.core.body as body
import ansys.speos.core.face as face
import ansys.speos.core.generic.general_methods as general_methods
from ansys.speos.core.generic.visualization_methods import _VisualArrow, _VisualData
from ansys.speos.core.geo_ref import GeoRef
from ansys.speos.core.intensity import Intensity
from ansys.speos.core.kernel.client import SpeosClient
from ansys.speos.core.kernel.scene import ProtoScene
from ansys.speos.core.kernel.source_template import ProtoSourceTemplate
from ansys.speos.core.spectrum import Spectrum


class BaseSource:
    """
    Super Class for all sources.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project in which source shall be created.
    name : str
        Name of the source.
    description : str
        Description of the source.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    source_instance : ansys.api.speos.scene.v2.scene_pb2.Scene.SourceInstance, optional
        Source instance to provide if the feature does not have to be created from scratch
        By default, ``None``, means that the feature is created from scratch by default.

    Notes
    -----
    This is a Super class, **Do not instantiate this class yourself**
    """

    class UserDefinedColorSpace:
        """Type of color space is user defined.

        Parameters
        ----------
        userdefined_color_space : source_pb2.SourceTemplate.UserDefinedRGBSpace
            source_pb2.SourceTemplate.UserDefinedRGBSpace
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_userdefined_color_space() method.
        """

        class UserDefinedWhitePoint:
            """Type of white point is user defined.

            Parameters
            ----------
            userdefined_white_point : source_pb2.SourceTemplate.UserDefinedWhitePoint
                source_pb2.SourceTemplate.UserDefinedWhitePoint
            default_values : bool
                Uses default values when True.
            stable_ctr : bool
                Variable to indicate if usage is inside class scope

            Notes
            -----
            **Do not instantiate this class yourself**,
            use set_white_point_type_user_defined() method.
            """

            def __init__(
                self,
                userdefined_white_point: source_pb2.SourceTemplate.UserDefinedWhitePoint,
                default_values: bool = True,
                stable_ctr: bool = True,
            ):
                if not stable_ctr:
                    msg = "UserDefinedWhitePoint class instantiated outside of class scope"
                    raise RuntimeError(msg)
                self._userdefined_white_point = userdefined_white_point

                if default_values:
                    self.white_point = [0.31271, 0.32902]

            @property
            def white_point(self):
                """White point coordinate.

                This property gets or sets the white point coordinate [x, y]

                Parameters
                ----------
                value: List[float]
                    The white point coordinate, [0.31271, 0.32902] by default.

                Returns
                -------
                List[float]
                    User defined white point coordinate

                """
                return self._userdefined_white_point.white_point

            @white_point.setter
            def white_point(self, value: List[float]):
                self._userdefined_white_point.white_point[:] = value

        def __init__(
            self,
            project: project.Project,
            userdefined_color_space: source_pb2.SourceTemplate.UserDefinedRGBSpace,
            default_values: bool = True,
            stable_ctr: bool = True,
        ):
            self._project = project
            if not stable_ctr:
                msg = "UserDefinedColorSpace class instantiated outside of class scope"
                raise RuntimeError(msg)
            self._userdefined_color_space = userdefined_color_space
            self._white_point_type = None

            self._red_spectrum = BaseSource._Spectrum(
                speos_client=self._project.client,
                name="",
                message_to_complete=self._userdefined_color_space,
                field_name_to_complete="red_spectrum_guid",
                spectrum_guid=self._userdefined_color_space.red_spectrum_guid,
            )

            self._green_spectrum = BaseSource._Spectrum(
                speos_client=self._project.client,
                name="",
                message_to_complete=self._userdefined_color_space,
                field_name_to_complete="green_spectrum_guid",
                spectrum_guid=self._userdefined_color_space.green_spectrum_guid,
            )

            self._blue_spectrum = BaseSource._Spectrum(
                speos_client=self._project.client,
                name="",
                message_to_complete=self._userdefined_color_space,
                field_name_to_complete="blue_spectrum_guid",
                spectrum_guid=self._userdefined_color_space.blue_spectrum_guid,
            )

            if default_values:
                # Default values
                self.set_white_point_type_d65()

        @property
        def red_spectrum(self) -> dict:
            """Get red spectrum.

            Returns
            -------
            dict
                Red spectrum dictionary

            """
            return self._red_spectrum._spectrum._to_dict()

        @red_spectrum.setter
        def red_spectrum(self, red_spectrum_file_uri: str) -> None:
            """Set red spectrum.

            Parameters
            ----------
            red_spectrum_file_uri: str
                Red spectrum file uri.

            Returns
            -------
            None
            """
            if self._red_spectrum._message_to_complete is not self._userdefined_color_space:
                # Happens in case of feature reset (to be sure to always modify correct data)
                self._red_spectrum._message_to_complete = self._userdefined_color_space

            # name for the spectrum chosen: <file_uri>.Spectrum
            self._red_spectrum._spectrum._spectrum.name = Path(red_spectrum_file_uri).name
            self._red_spectrum._spectrum.set_library(file_uri=red_spectrum_file_uri)

        @property
        def green_spectrum(self) -> dict:
            """Get green spectrum.

            Returns
            -------
            dict
                Green spectrum dictionary

            """
            return self._green_spectrum._spectrum._to_dict()

        @green_spectrum.setter
        def green_spectrum(self, green_spectrum_file_uri: str) -> None:
            """Set green spectrum.

            Parameters
            ----------
            green_spectrum_file_uri: str
                Green spectrum file uri.

            Returns
            -------
            None
            """
            if self._green_spectrum._message_to_complete is not self._userdefined_color_space:
                # Happens in case of feature reset (to be sure to always modify correct data)
                self._green_spectrum._message_to_complete = self._userdefined_color_space

            # name for the spectrum chosen: <file_uri>.Spectrum
            self._green_spectrum._spectrum._spectrum.name = Path(green_spectrum_file_uri).name
            self._green_spectrum._spectrum.set_library(file_uri=green_spectrum_file_uri)

        @property
        def blue_spectrum(self) -> dict:
            """Get blue spectrum.

            Returns
            -------
            dict
                Blue spectrum dictionary

            """
            return self._blue_spectrum._spectrum._to_dict()

        @blue_spectrum.setter
        def blue_spectrum(self, blue_spectrum_file_uri: str) -> None:
            """Set blue spectrum.

            Parameters
            ----------
            blue_spectrum_file_uri: str
                Blue spectrum file uri.

            Returns
            -------
            None
            """
            if self._blue_spectrum._message_to_complete is not self._userdefined_color_space:
                # Happens in case of feature reset (to be sure to always modify correct data)
                self._blue_spectrum._message_to_complete = self._userdefined_color_space

            # name for the spectrum chosen: <file_uri>.Spectrum
            self._blue_spectrum._spectrum._spectrum.name = Path(blue_spectrum_file_uri).name
            self._blue_spectrum._spectrum.set_library(file_uri=blue_spectrum_file_uri)

        @property
        def white_point_type(
            self,
        ) -> Union[
            None,
            source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.D65,
            source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.D50,
            source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.C,
            source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.E,
            UserDefinedWhitePoint,
        ]:
            """Get the white point type.

            Returns
            -------
            Union
            [
                None,
                source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.D65,
                source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.D50,
                source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.C,
                source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.E,
                UserDefinedWhitePoint,
            ]
            Predefined White Point Type or User Defined White Point Type.

            """
            return self._white_point_type

        def set_white_point_type_d65(self) -> None:
            """Set white point type to D65.

            Returns
            -------
            None

            """
            self._userdefined_color_space.pre_defined_white_point.white_point_type = (
                source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.D65
            )
            self._white_point_type = (
                source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.D65
            )

        def set_white_point_type_c(self) -> None:
            """Set white point type to C.

            Returns
            -------
            None

            """
            self._userdefined_color_space.pre_defined_white_point.white_point_type = (
                source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.C
            )
            self._white_point_type = source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.C

        def set_white_point_type_d50(self) -> None:
            """Set white point type to D50.

            Returns
            -------
            None

            """
            self._userdefined_color_space.pre_defined_white_point.white_point_type = (
                source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.D50
            )
            self._white_point_type = (
                source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.D50
            )

        def set_white_point_type_e(self) -> None:
            """Set white point type to E.

            Returns
            -------
            None

            """
            self._userdefined_color_space.pre_defined_white_point.white_point_type = (
                source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.E
            )
            self._white_point_type = source_pb2.SourceTemplate.PredefinedWhitePoint.WhitePointType.E

        def set_white_point_type_user_defined(self) -> UserDefinedWhitePoint:
            """Set white point type to user_defined.

            Returns
            -------
            UserDefinedWhitePoint
                User defined white point settings.

            """
            if self._white_point_type is None and self._userdefined_color_space.HasField(
                "user_defined_white_point"
            ):
                self._white_point_type = BaseSource.UserDefinedColorSpace.UserDefinedWhitePoint(
                    userdefined_white_point=self._userdefined_color_space.user_defined_white_point,
                    default_values=False,
                    stable_ctr=True,
                )
            if not isinstance(
                self._white_point_type, BaseSource.UserDefinedColorSpace.UserDefinedWhitePoint
            ):
                # if the _type is not UserDefinedWhitePoint then we create a new type.
                self._white_point_type = BaseSource.UserDefinedColorSpace.UserDefinedWhitePoint(
                    userdefined_white_point=self._userdefined_color_space.user_defined_white_point,
                    stable_ctr=True,
                )
            elif (
                self._white_point_type._userdefined_white_point
                is not self._userdefined_color_space.user_defined_white_point
            ):
                # Happens in case of feature reset (to be sure to always modify correct data)
                self._white_point_type._userdefined_white_point = (
                    self._userdefined_color_space.user_defined_white_point
                )
            return self._white_point_type

    class PredefinedColorSpace:
        """Type of color space is predefined value.

        Parameters
        ----------
        predefined_color_space :
            ansys.api.speos.source.v1.source_pb2.SourceTemplate.PredefinedColorSpace
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_predefined_color_space() method.
        """

        def __init__(
            self,
            predefined_color_space: source_pb2.SourceTemplate.PredefinedColorSpace,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "PredefinedColorSpace class instantiated outside of class scope"
                raise RuntimeError(msg)
            self._predefined_color_space = predefined_color_space

            if default_values:
                # Default values
                self.set_color_space_srgb()

        def set_color_space_srgb(self) -> SourceAmbientEnvironment.PredefinedColorSpace:
            """Set the color space to the srgb preset.

            Returns
            -------
            ansys.speos.core.source.SourceAmbientEnvironment.PredefinedColorSpace
            """
            self._predefined_color_space.color_space_type = (
                source_pb2.SourceTemplate.PredefinedColorSpace.sRGB
            )
            return self

        def set_color_space_adobergb(self) -> SourceAmbientEnvironment.PredefinedColorSpace:
            """Set the color space to the Adobe RGB preset.

            Returns
            -------
            ansys.speos.core.source.SourceAmbientEnvironment.PredefinedColorSpace
            """
            self._predefined_color_space.color_space_type = (
                source_pb2.SourceTemplate.PredefinedColorSpace.AdobeRGB
            )
            return self

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
    ) -> None:
        self._project = project
        self._name = name
        self._unique_id = None
        self._visual_data = _VisualData(ray=True) if general_methods._GRAPHICS_AVAILABLE else None
        self.source_template_link = None
        """Link object for the source template in database."""

        if metadata is None:
            metadata = {}

        if source_instance is None:
            # Create local SourceTemplate
            self._source_template = ProtoSourceTemplate(
                name=name, description=description, metadata=metadata
            )

            # Create local SourceInstance
            self._source_instance = ProtoScene.SourceInstance(
                name=name, description=description, metadata=metadata
            )
        else:
            self._unique_id = source_instance.metadata["UniqueId"]
            self.source_template_link = self._project.client[source_instance.source_guid]
            self._reset()

    class _Spectrum:
        def __init__(
            self,
            speos_client: SpeosClient,
            name: str,
            message_to_complete: Union[
                ProtoSourceTemplate.RayFile,
                ProtoSourceTemplate.Surface,
                ProtoSourceTemplate.Luminaire,
            ],
            field_name_to_complete="",
            spectrum_guid: str = "",
        ) -> None:
            self._message_to_complete = message_to_complete
            self._field_name_to_complete = field_name_to_complete
            if spectrum_guid != "":
                self._spectrum = Spectrum(
                    speos_client=speos_client,
                    name=name + ".Spectrum",
                    key=spectrum_guid,
                )
            else:
                self._spectrum = Spectrum(speos_client=speos_client, name=name + ".Spectrum")

            self._no_spectrum = None  # None means never committed, or deleted
            self._no_spectrum_local = False

        def __str__(self) -> str:
            if self._no_spectrum is None:
                if self._no_spectrum_local is False:
                    return str(self._spectrum)
            else:
                if self._no_spectrum is False:
                    return str(self._spectrum)
            return ""

        def _commit(self) -> BaseSource._Spectrum:
            if not self._no_spectrum_local:
                self._spectrum.commit()
                if self._field_name_to_complete == "":
                    self._message_to_complete.spectrum_guid = self._spectrum.spectrum_link.key
                elif self._field_name_to_complete == "red_spectrum_guid":
                    self._message_to_complete.red_spectrum_guid = self._spectrum.spectrum_link.key
                elif self._field_name_to_complete == "green_spectrum_guid":
                    self._message_to_complete.green_spectrum_guid = self._spectrum.spectrum_link.key
                elif self._field_name_to_complete == "blue_spectrum_guid":
                    self._message_to_complete.blue_spectrum_guid = self._spectrum.spectrum_link.key
                self._no_spectrum = self._no_spectrum_local
            return self

        def _reset(self) -> BaseSource._Spectrum:
            self._spectrum.reset()
            if self._no_spectrum is not None:
                self._no_spectrum_local = self._no_spectrum
            return self

        def _delete(self) -> BaseSource._Spectrum:
            self._no_spectrum = None
            return self

    def _to_dict(self) -> dict:
        out_dict = {}

        # SourceInstance (= source guid + source properties)
        if self._project.scene_link and self._unique_id is not None:
            scene_data = self._project.scene_link.get()
            src_inst = next(
                (x for x in scene_data.sources if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if src_inst is not None:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._project.client, message=src_inst
                )
            else:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self._source_instance,
                )
        else:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._project.client, message=self._source_instance
            )

        if "source" not in out_dict.keys():
            # SourceTemplate
            if self.source_template_link is None:
                out_dict["source"] = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self._source_template,
                )
            else:
                out_dict["source"] = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self.source_template_link.get(),
                )

        # # handle spectrum & intensity
        # if self._type is not None:
        #     self._type._to_dict(dict_to_complete=out_dict)

        proto_message_utils._replace_properties(json_dict=out_dict)

        return out_dict

    def get(self, key: str = "") -> list[tuple[str, dict]]:
        """Get dictionary corresponding to the project - read only.

        Parameters
        ----------
        key: str

        Returns
        -------
        str | dict
        """
        if key == "":
            return self._to_dict()
        info = proto_message_utils._value_finder_key_startswith(dict_var=self._to_dict(), key=key)
        content = list(info)
        if len(content) != 0:
            content.sort(
                key=lambda x: SequenceMatcher(None, x[0], key).ratio(),
                reverse=True,
            )
            return content[0][1]
        info = proto_message_utils._flatten_dict(dict_var=self._to_dict())
        print("Used key: {} not found in key list: {}.".format(key, info.keys()))

    def __str__(self) -> str:
        """Return the string representation of the source."""
        out_str = ""
        if self._project.scene_link and self._unique_id is not None:
            scene_data = self._project.scene_link.get()
            src_inst = next(
                (x for x in scene_data.sources if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if src_inst is None:
                out_str += "local: "
        else:
            out_str += "local: "

        out_str += proto_message_utils.dict_to_str(dict=self._to_dict())
        return out_str

    def _commit(self) -> BaseSource:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        # The _unique_id will help to find correct item in the scene.sources:
        # the list of SourceInstance
        if self._unique_id is None:
            self._unique_id = str(uuid.uuid4())
            self._source_instance.metadata["UniqueId"] = self._unique_id

        # Save or Update the source template (depending on if it was already saved before)
        if self.source_template_link is None:
            self.source_template_link = self._project.client.source_templates().create(
                message=self._source_template
            )
            self._source_instance.source_guid = self.source_template_link.key
        elif self.source_template_link.get() != self._source_template:
            self.source_template_link.set(
                data=self._source_template
            )  # Only update if template has changed

        # Update the scene with the source instance
        if self._project.scene_link:
            update_scene = True
            scene_data = self._project.scene_link.get()  # retrieve scene data

            # Look if an element corresponds to the _unique_id
            src_inst = next(
                (x for x in scene_data.sources if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if src_inst is not None:
                if src_inst != self._source_instance:
                    src_inst.CopyFrom(self._source_instance)  # if yes, just replace
                else:
                    update_scene = False
            else:
                scene_data.sources.append(
                    self._source_instance
                )  # if no, just add it to the list of sources

            if update_scene:  # Update scene only if instance has changed
                self._project.scene_link.set(data=scene_data)  # update scene data

        return self

    def _reset(self) -> BaseSource:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        # Reset source template
        if self.source_template_link is not None:
            self._source_template = self.source_template_link.get()

        # Reset source instance
        if self._project.scene_link is not None:
            scene_data = self._project.scene_link.get()  # retrieve scene data
            # Look if an element corresponds to the _unique_id
            src_inst = next(
                (x for x in scene_data.sources if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if src_inst is not None:
                self._source_instance = src_inst
        return self

    def _delete(self) -> BaseSource:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        # This allows to clean-managed object contained in _luminaire, _rayfile, etc..
        # Like Spectrum, IntensityTemplate

        # Delete the source template
        if self.source_template_link is not None:
            self.source_template_link.delete()
            self.source_template_link = None

        # Reset then the source_guid (as the source template was deleted just above)
        self._source_instance.source_guid = ""

        # Remove the source from the scene
        scene_data = self._project.scene_link.get()  # retrieve scene data
        src_inst = next(
            (x for x in scene_data.sources if x.metadata["UniqueId"] == self._unique_id),
            None,
        )
        if src_inst is not None:
            scene_data.sources.remove(src_inst)
            self._project.scene_link.set(data=scene_data)  # update scene data

        # Reset the _unique_id
        self._unique_id = None
        self._source_instance.metadata.pop("UniqueId")
        return self

    def _fill(self, src_inst):
        self._unique_id = src_inst.metadata["UniqueId"]
        self._source_instance = src_inst
        self.source_template_link = self._project.client[src_inst.source_guid]
        self._reset()

    def commit(self) -> BaseSource:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        if hasattr(self, "_spectrum"):
            self._spectrum._commit()
        self._commit()
        if general_methods._GRAPHICS_AVAILABLE:
            self._visual_data.updated = False
        return self

    def reset(self) -> BaseSource:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        if hasattr(self, "_spectrum"):
            self._spectrum._reset()
        self._reset()
        return self

    def delete(self) -> BaseSource:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.source.BaseSource
            Source feature.
        """
        if hasattr(self, "_spectrum"):
            self._spectrum._delete()
        self._delete()
        return self


class SourceLuminaire(BaseSource):
    """LuminaireSource.

    By default, a flux from intensity file is chosen, with an incandescent spectrum.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    default_values : bool
        Uses default values when True.
    """

    @general_methods.min_speos_version(25, 2, 0)
    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            source_instance=source_instance,
        )

        self._spectrum = self._Spectrum(
            speos_client=self._project.client,
            name=name,
            message_to_complete=self._source_template.luminaire,
            spectrum_guid=self._source_template.luminaire.spectrum_guid,
        )

        if default_values:
            # Default values
            self.set_flux_from_intensity_file().set_spectrum().set_incandescent()
            self.set_axis_system()

    @property
    def visual_data(self) -> _VisualData:
        """Property containing Luminaire source visualization data.

        Returns
        -------
        _VisualData
            Instance of VisualData Class for pyvista.PolyData of feature rays, coordinate_systems.

        """
        if self._visual_data.updated:
            return self._visual_data
        else:
            self._visual_data = (
                _VisualData(ray=True) if general_methods._GRAPHICS_AVAILABLE else None
            )
            for ray_path in self._project.scene_link.get_source_ray_paths(
                self._name, rays_nb=100, raw_data=True, display_data=True
            ):
                self._visual_data.add_data_line(
                    _VisualArrow(
                        line_vertices=[ray_path.impacts_coordinates, ray_path.last_direction],
                        color=tuple(ray_path.colors.values),
                        arrow=False,
                    )
                )
            feature_pos_info = self.get(key="axis_system")
            feature_luminaire_pos = np.array(feature_pos_info[:3])
            feature_luminaire_x_dir = np.array(feature_pos_info[3:6])
            feature_luminaire_y_dir = np.array(feature_pos_info[6:9])
            feature_luminaire_z_dir = np.array(feature_pos_info[9:12])
            self._visual_data.coordinates.origin = feature_luminaire_pos
            self._visual_data.coordinates.x_axis = feature_luminaire_x_dir
            self._visual_data.coordinates.y_axis = feature_luminaire_y_dir
            self._visual_data.coordinates.z_axis = feature_luminaire_z_dir
            self._visual_data.updated = True
            return self._visual_data

    def set_flux_from_intensity_file(self) -> SourceLuminaire:
        """Take flux from intensity file provided.

        Returns
        -------
        ansys.speos.core.source.SourceLuminaire
            Luminaire source.
        """
        self._source_template.luminaire.flux_from_intensity_file.SetInParent()
        return self

    def set_flux_luminous(self, value: float = 683) -> SourceLuminaire:
        """Set luminous flux.

        Parameters
        ----------
        value : float
            Luminous flux in lumens.
            By default, ``683.0``.

        Returns
        -------
        ansys.speos.core.source.SourceLuminaire
            Luminaire source.
        """
        self._source_template.luminaire.luminous_flux.luminous_value = value
        return self

    def set_flux_radiant(self, value: float = 1) -> SourceLuminaire:
        """Set radiant flux.

        Parameters
        ----------
        value : float
            Radiant flux in watts.
            By default, ``1.0``.

        Returns
        -------
        ansys.speos.core.source.SourceLuminaire
            Luminaire source.
        """
        self._source_template.luminaire.radiant_flux.radiant_value = value
        return self

    def set_intensity_file_uri(self, uri: str) -> SourceLuminaire:
        """Set intensity file.

        Parameters
        ----------
        uri : str
            IES or EULUMDAT format file uri.

        Returns
        -------
        ansys.speos.core.source.SourceLuminaire
            Luminaire source.
        """
        self._source_template.luminaire.intensity_file_uri = uri
        return self

    def set_spectrum(self) -> Spectrum:
        """Set spectrum.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum.
        """
        if self._spectrum._message_to_complete is not self._source_template.luminaire:
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._spectrum._message_to_complete = self._source_template.luminaire
        return self._spectrum._spectrum

    def set_axis_system(self, axis_system: Optional[List[float]] = None) -> SourceLuminaire:
        """Set the position of the source.

        Parameters
        ----------
        axis_system : Optional[List[float]]
            Position of the source [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz].
            By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]``.

        Returns
        -------
        ansys.speos.core.source.SourceLuminaire
            Luminaire source.
        """
        if axis_system is None:
            axis_system = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]
        self._source_instance.luminaire_properties.axis_system[:] = axis_system
        return self


class SourceRayFile(BaseSource):
    """RayFile Source.

    By default, flux and spectrum from ray file are selected.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    default_values : bool
        Uses default values when True.
    """

    @general_methods.min_speos_version(25, 2, 0)
    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            source_instance=source_instance,
        )
        self._client = self._project.client

        spectrum_guid = ""
        if self._source_template.rayfile.HasField("spectrum_guid"):
            spectrum_guid = self._source_template.rayfile.spectrum_guid
        self._spectrum = self._Spectrum(
            speos_client=self._client,
            name=name,
            message_to_complete=self._source_template.rayfile,
            spectrum_guid=spectrum_guid,
        )
        if spectrum_guid == "":
            self.set_spectrum_from_ray_file()

        self._name = name

        if default_values:
            # Default values
            self.set_flux_from_ray_file().set_spectrum_from_ray_file()
            self.set_axis_system()

    @property
    def visual_data(self) -> _VisualData:
        """Property containing Rayfile source visualization data.

        Returns
        -------
        _VisualData
            Instance of VisualData Class for pyvista.PolyData of feature rays, coordinate_systems.

        """
        if self._visual_data.updated:
            return self._visual_data
        else:
            self._visual_data = (
                _VisualData(ray=True) if general_methods._GRAPHICS_AVAILABLE else None
            )
            for ray_path in self._project.scene_link.get_source_ray_paths(
                self._name, rays_nb=100, raw_data=True, display_data=True
            ):
                self._visual_data.add_data_line(
                    _VisualArrow(
                        line_vertices=[ray_path.impacts_coordinates, ray_path.last_direction],
                        color=tuple(ray_path.colors.values),
                        arrow=False,
                    )
                )
            feature_pos_info = self.get(key="axis_system")
            feature_rayfile_pos = np.array(feature_pos_info[:3])
            feature_rayfile_x_dir = np.array(feature_pos_info[3:6])
            feature_rayfile_y_dir = np.array(feature_pos_info[6:9])
            feature_rayfile_z_dir = np.array(feature_pos_info[9:12])
            self._visual_data.coordinates.origin = feature_rayfile_pos
            self._visual_data.coordinates.x_axis = feature_rayfile_x_dir
            self._visual_data.coordinates.y_axis = feature_rayfile_y_dir
            self._visual_data.coordinates.z_axis = feature_rayfile_z_dir
            self._visual_data.updated = True
            return self._visual_data

    def set_ray_file_uri(self, uri: str) -> SourceRayFile:
        """Set ray file.

        Parameters
        ----------
        uri : str
            Rayfile format file uri (.ray or .tm25ray files expected).

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile source.
        """
        self._source_template.rayfile.ray_file_uri = uri
        return self

    def set_flux_from_ray_file(self) -> SourceRayFile:
        """Take flux from ray file provided.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile source.
        """
        self._source_template.rayfile.flux_from_ray_file.SetInParent()
        return self

    def set_flux_luminous(self, value: float = 683) -> SourceRayFile:
        """Set luminous flux.

        Parameters
        ----------
        value : float
            Luminous flux in lumens.
            By default, ``683.0``.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile source.
        """
        self._source_template.rayfile.luminous_flux.luminous_value = value
        return self

    def set_flux_radiant(self, value: float = 1) -> SourceRayFile:
        """Set radiant flux.

        Parameters
        ----------
        value : float
            Radiant flux in watts.
            By default, ``1.0``.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile source.
        """
        self._source_template.rayfile.radiant_flux.radiant_value = value
        return self

    def set_spectrum_from_ray_file(self) -> SourceRayFile:
        """Take spectrum from ray file provided.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile source.
        """
        self._source_template.rayfile.spectrum_from_ray_file.SetInParent()
        self._spectrum._no_spectrum_local = True
        return self

    def set_spectrum(self) -> Spectrum:
        """Set spectrum of the Source.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum.
        """
        if self._source_template.rayfile.HasField("spectrum_from_ray_file"):
            guid = ""
            if self._spectrum._spectrum.spectrum_link is not None:
                guid = self._spectrum._spectrum.spectrum_link.key
            self._source_template.rayfile.spectrum_guid = guid

        if self._spectrum._message_to_complete is not self._source_template.rayfile:
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._spectrum._message_to_complete = self._source_template.rayfile

        self._spectrum._no_spectrum_local = False
        return self._spectrum._spectrum

    def set_axis_system(self, axis_system: Optional[List[float]] = None) -> SourceRayFile:
        """Set position of the source.

        Parameters
        ----------
        axis_system : Optional[List[float]]
            Position of the source [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz].
            By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]``.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile Source.
        """
        if axis_system is None:
            axis_system = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]
        self._source_instance.rayfile_properties.axis_system[:] = axis_system
        return self

    def set_exit_geometries(self, exit_geometries: Optional[List[GeoRef]] = None) -> SourceRayFile:
        """Set exit geometries.

        Parameters
        ----------
        exit_geometries : List[ansys.speos.core.geo_ref.GeoRef]
            Exit Geometries that will use this rayfile source.
            By default, ``[]``.

        Returns
        -------
        ansys.speos.core.source.SourceRayFile
            RayFile Source.
        """
        if not exit_geometries:
            self._source_instance.rayfile_properties.ClearField("exit_geometries")
        else:
            self._source_instance.rayfile_properties.exit_geometries.geo_paths[:] = [
                gr.to_native_link() for gr in exit_geometries
            ]

        return self


class SourceSurface(BaseSource):
    """Type of Source : Surface.

    By default, a luminous flux and existence constant are chosen. With a monochromatic spectrum,
    and lambertian intensity (cos with N = 1).

    Parameters
    ----------
    speos_client : ansys.speos.core.kernel.client.SpeosClient
        The Speos instance client.
    name : str
        Name of the source feature.
    surface : ansys.api.speos.source.v1.source_pb2.SourceTemplate.Surface
        Surface source to complete.
    surface_props : ansys.api.speos.scene.v2.scene_pb2.Scene.SourceInstance.SurfaceProperties
        Surface source properties to complete.
    default_values : bool
        Uses default values when True.
    """

    class ExitanceVariable:
        """Type of surface source existence : existence variable.

        Parameters
        ----------
        exitance_variable : ansys.api.speos.source.v1.source_pb2.SourceTemplate.Surface.
        ExitanceVariable
            Existence variable to complete.
        exitance_variable_props : ansys.api.speos.scene.v2.scene_pb2.Scene.SourceInstance.
        SurfaceProperties.ExitanceVariableProperties
            Existence variable properties to complete.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_exitance_variable method available in
        Source classes.
        """

        def __init__(
            self,
            exitance_variable,
            exitance_variable_props,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "ExitanceVariable class instantiated outside of class scope"
                raise RuntimeError(msg)
            self._exitance_variable = exitance_variable
            self._exitance_variable_props = exitance_variable_props

            if default_values:
                # Default values
                self._exitance_variable.SetInParent()
                self.set_axis_plane()

        def set_xmp_file_uri(self, uri: str) -> SourceSurface.ExitanceVariable:
            """Set existence xmp file.

            Parameters
            ----------
            uri : str
                XMP file describing existence.

            Returns
            -------
            ansys.speos.core.source.SourceSurface.ExitanceVariable
                ExitanceVariable of surface source.
            """
            self._exitance_variable.exitance_xmp_file_uri = uri
            return self

        def set_axis_plane(
            self, axis_plane: Optional[List[float]] = None
        ) -> SourceSurface.ExitanceVariable:
            """Set position of the existence map.

            Parameters
            ----------
            axis_plane : Optional[List[float]]
                Position of the existence map [Ox Oy Oz Xx Xy Xz Yx Yy Yz].
                By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0]``.

            Returns
            -------
            ansys.speos.core.source.SourceSurface.ExitanceVariable
                ExitanceVariable of surface Source.
            """
            if axis_plane is None:
                axis_plane = [0, 0, 0, 1, 0, 0, 0, 1, 0]
            self._exitance_variable_props.axis_plane[:] = axis_plane
            return self

    @general_methods.min_speos_version(25, 2, 0)
    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            source_instance=source_instance,
        )
        self._speos_client = self._project.client
        self._name = name

        spectrum_guid = ""
        if self._source_template.surface.HasField("spectrum_guid"):
            spectrum_guid = self._source_template.surface.spectrum_guid
        self._spectrum = self._Spectrum(
            speos_client=self._speos_client,
            name=name,
            message_to_complete=self._source_template.surface,
            spectrum_guid=spectrum_guid,
        )

        self._intensity = Intensity(
            speos_client=self._speos_client,
            name=name + ".Intensity",
            intensity_props_to_complete=self._source_instance.surface_properties.intensity_properties,
            key=self._source_template.surface.intensity_guid,
        )

        # Attribute gathering more complex existence type
        self._exitance_type = None

        if default_values:
            # Default values
            self.set_flux_luminous().set_exitance_constant(geometries=[]).set_intensity()
            self.set_spectrum()

    @property
    def visual_data(self) -> _VisualData:
        """Property containing Surface source visualization data.

        Returns
        -------
        _VisualData
            Instance of VisualData Class for pyvista.PolyData of feature rays, coordinate_systems.

        """
        if self._visual_data.updated:
            return self._visual_data
        else:
            self._visual_data = (
                _VisualData(
                    ray=True,
                    coordinate_system=True if self._exitance_type is not None else False,
                )
                if general_methods._GRAPHICS_AVAILABLE
                else None
            )
            for ray_path in self._project.scene_link.get_source_ray_paths(
                self._name, rays_nb=100, raw_data=True, display_data=True
            ):
                self._visual_data.add_data_line(
                    _VisualArrow(
                        line_vertices=[ray_path.impacts_coordinates, ray_path.last_direction],
                        color=tuple(ray_path.colors.values),
                        arrow=False,
                    )
                )
            if self._visual_data.coordinates is not None:
                feature_pos_info = self.get(key="axis_plane")
                feature_surface_pos = np.array(feature_pos_info[:3])
                feature_surface_x_dir = np.array(feature_pos_info[3:6])
                feature_surface_y_dir = np.array(feature_pos_info[6:9])
                self._visual_data.coordinates.origin = feature_surface_pos
                self._visual_data.coordinates.x_axis = feature_surface_x_dir
                self._visual_data.coordinates.y_axis = feature_surface_y_dir
            self._visual_data.updated = True
            return self._visual_data

    def set_flux_from_intensity_file(self) -> SourceSurface:
        """Take flux from intensity file provided.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._source_template.surface.flux_from_intensity_file.SetInParent()
        return self

    def set_flux_luminous(self, value: float = 683) -> SourceSurface:
        """Set luminous flux.

        Parameters
        ----------
        value : float
            Luminous flux in lumens.
            By default, ``683.0``.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._source_template.surface.luminous_flux.luminous_value = value
        return self

    def set_flux_radiant(self, value: float = 1) -> SourceSurface:
        """Set radiant flux.

        Parameters
        ----------
        value : float
            Radiant flux in watts.
            By default, ``1.0``.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._source_template.surface.radiant_flux.radiant_value = value
        return self

    def set_flux_luminous_intensity(self, value: float = 5) -> SourceSurface:
        """Set luminous intensity flux.

        Parameters
        ----------
        value : float
            Luminous intensity in candelas.
            By default, ``5.0``.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._source_template.surface.luminous_intensity_flux.luminous_intensity_value = value
        return self

    def set_intensity(self) -> Intensity:
        """Set intensity.

        Returns
        -------
        ansys.speos.core.intensity.Intensity
            Intensity.
        """
        if (
            self._intensity._intensity_properties
            is not self._source_instance.surface_properties.intensity_properties
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._intensity._intensity_properties = (
                self._source_instance.surface_properties.intensity_properties
            )

        return self._intensity

    def set_exitance_constant(
        self, geometries: List[tuple[Union[GeoRef, face.Face, body.Body], bool]]
    ) -> SourceSurface:
        """Set existence constant.

        Parameters
        ----------
        geometries : List[tuple[ansys.speos.core.geo_ref.GeoRef, bool]]
            List of (face, reverseNormal).

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._exitance_type = None

        self._source_template.surface.exitance_constant.SetInParent()
        self._source_instance.surface_properties.exitance_constant_properties.ClearField(
            "geo_paths"
        )
        if geometries != []:
            geo_paths = []
            for gr, reverse_normal in geometries:
                if isinstance(gr, GeoRef):
                    geo_paths.append(
                        ProtoScene.GeoPath(
                            geo_path=gr.to_native_link(), reverse_normal=reverse_normal
                        )
                    )
                elif isinstance(gr, (face.Face, body.Body)):
                    geo_paths.append(
                        ProtoScene.GeoPath(
                            geo_path=gr.geo_path.to_native_link(), reverse_normal=reverse_normal
                        )
                    )
                else:
                    msg = f"Type {type(gr)} is not supported as Surface Source geometry input."
                    raise TypeError(msg)
            self._source_instance.surface_properties.exitance_constant_properties.geo_paths.extend(
                geo_paths
            )
        return self

    def set_exitance_variable(self) -> SourceSurface.ExitanceVariable:
        """Set existence variable, taken from XMP map.

        Returns
        -------
        ansys.speos.core.source.SourceSurface.ExitanceVariable
            ExitanceVariable of surface source.
        """
        if self._exitance_type is None and self._source_template.surface.HasField(
            "exitance_variable"
        ):
            # Happens in case of project created via load of speos file
            self._exitance_type = SourceSurface.ExitanceVariable(
                exitance_variable=self._source_template.surface.exitance_variable,
                exitance_variable_props=self._source_instance.surface_properties.exitance_variable_properties,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._exitance_type, SourceSurface.ExitanceVariable):
            # if the _exitance_type is not ExitanceVariable then we create a new type.
            self._exitance_type = SourceSurface.ExitanceVariable(
                exitance_variable=self._source_template.surface.exitance_variable,
                exitance_variable_props=self._source_instance.surface_properties.exitance_variable_properties,
                stable_ctr=True,
            )
        elif (
            self._exitance_type._exitance_variable
            is not self._source_template.surface.exitance_variable
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._exitance_type._exitance_variable = self._source_template.surface.exitance_variable
            self._exitance_type._exitance_variable_props = (
                self._source_instance.surface_properties.exitance_variable_properties
            )
        return self._exitance_type

    def set_spectrum_from_xmp_file(self) -> SourceSurface:
        """Take spectrum from xmp file provided.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Surface source.
        """
        self._source_template.surface.spectrum_from_xmp_file.SetInParent()
        self._spectrum._no_spectrum_local = True
        return self

    def set_spectrum(self) -> Spectrum:
        """Set spectrum of the Source.

        Returns
        -------
        ansys.speos.core.spectrum.Spectrum
            Spectrum.
        """
        if self._source_template.surface.HasField("spectrum_from_xmp_file"):
            guid = ""
            if self._spectrum._spectrum.spectrum_link is not None:
                guid = self._spectrum._spectrum.spectrum_link.key
            self._source_template.surface.spectrum_guid = guid

        if self._spectrum._message_to_complete is not self._source_template.surface:
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._spectrum._message_to_complete = self._source_template.surface

        self._spectrum._no_spectrum_local = False
        return self._spectrum._spectrum

    def commit(self) -> SourceSurface:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Source feature.
        """
        # intensity
        self._intensity.commit()
        self._source_template.surface.intensity_guid = self._intensity.intensity_template_link.key

        # spectrum & source
        super().commit()
        return self

    def reset(self) -> SourceSurface:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Source feature.
        """
        self._intensity.reset()
        # spectrum & source
        super().reset()
        return self

    def delete(self) -> SourceSurface:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.source.SourceSurface
            Source feature.
        """
        # Currently we don't perform delete in cascade,
        # so deleting a surface source does not delete the intensity template used
        # self._intensity.delete()

        # spectrum & source
        super().delete()
        return self


class BaseSourceAmbient(BaseSource):
    """
    Super Class for ambient sources.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project in which source shall be created.
    name : str
        Name of the source.
    description : str
        Description of the source.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    source_instance : ansys.api.speos.scene.v2.scene_pb2.Scene.SourceInstance, optional
        Source instance to provide if the feature does not have to be created from scratch
        By default, ``None``, means that the feature is created from scratch by default.

    Notes
    -----
    This is a Super class, **Do not instantiate this class yourself**
    """

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
    ) -> None:
        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            source_instance=source_instance,
        )

    class AutomaticSun:
        """Sun type Automatic.

        By default, user's current time and Ansys France is used a time zone.

        Parameters
        ----------
        sun: ansys.api.speos.scene.v2.scene_pb2.AutomaticSun
            Wavelengths range protobuf object to modify.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_sun_automatic method available in
        source classes.
        """

        def __init__(
            self,
            sun: scene_pb2.AutomaticSun,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                raise RuntimeError(
                    "BaseSourceAmbient.AutomaticSun class instantiated outside of class scope"
                )
            self._sun = sun

            if default_values:
                now = datetime.datetime.now()
                self.year = now.year
                self.month = now.month
                self.day = now.day
                self.hour = now.hour
                self.minute = now.minute
                self.time_zone = "CET"
                self.longitude = 0.0
                self.latitude = 0.0

        @property
        def year(self) -> int:
            """Get year info of the automatic sun.

            Returns
            -------
            int
                year info.
            """
            return self._sun.year

        @year.setter
        def year(self, year: int) -> None:
            """Set year info of the automatic sun.

            Parameters
            ----------
            year: int
                year information.

            Returns
            -------
            None
            """
            self._sun.year = year

        @property
        def month(self) -> int:
            """Get month info of the automatic sun.

            Returns
            -------
            int
                month information.

            """
            return self._sun.month

        @month.setter
        def month(self, month: int) -> None:
            """Set month info of the automatic sun.

            Parameters
            ----------
            month: int
                month information.

            Returns
            -------
            None

            """
            self._sun.month = month

        @property
        def day(self) -> int:
            """Get day info of the automatic sun.

            Returns
            -------
            int
                day information.
            """
            return self._sun.day

        @day.setter
        def day(self, day: int) -> None:
            """Set day info of the automatic sun.

            Parameters
            ----------
            day: int
                day information.

            Returns
            -------
            None
            """
            self._sun.day = day

        @property
        def hour(self) -> int:
            """Get hour info of the automatic sun.

            Returns
            -------
            int
                hour information.

            """
            return self._sun.hour

        @hour.setter
        def hour(self, hour: int) -> None:
            """Set hour info of the automatic sun.

            Parameters
            ----------
            hour: int
                hour information.

            Returns
            -------
            None

            """
            self._sun.hour = hour

        @property
        def minute(self) -> int:
            """Get minute info of the automatic sun.

            Returns
            -------
            int
                minute information.

            """
            return self._sun.minute

        @minute.setter
        def minute(self, minute: int) -> None:
            """Set minute info of the automatic sun.

            Parameters
            ----------
            minute: int
                minute information.

            Returns
            -------
            None

            """
            self._sun.minute = minute

        @property
        def longitude(self) -> float:
            """Get longitude info of the automatic sun.

            Returns
            -------
            float
                longitude information.
            """
            return self._sun.longitude

        @longitude.setter
        def longitude(self, longitude: float) -> None:
            """Get longitude info of the automatic sun.

            Parameters
            ----------
            longitude: float
                longitude information.

            Returns
            -------
            None
            """
            self._sun.longitude = longitude

        @property
        def latitude(self) -> float:
            """Get latitude info of the automatic sun.

            Returns
            -------
            float
                latitude information.
            """
            return self._sun.latitude

        @latitude.setter
        def latitude(self, latitude: float) -> None:
            """Set latitude info of the automatic sun.

            Parameters
            ----------
            latitude: float
                latitude information.

            Returns
            -------
            None
            """
            self._sun.latitude = latitude

        @property
        def time_zone(self) -> str:
            """Get time zone info of the automatic sun.

            Returns
            -------
            str
                time zone abbreviation.
            """
            return self._sun.time_zone_uri

        @time_zone.setter
        def time_zone(self, time_zone: str) -> None:
            """Set time zone info of the automatic sun.

                default value to be "CET".

            Parameters
            ----------
            timezone: str
                timezone abbreviation.

            Returns
            -------
            None
            """
            self._sun.time_zone_uri = time_zone

    class Manual:
        """Sun type Manual>.

        By default, z-axis [0, 0, 1] is used as sun direction.

        Parameters
        ----------
        sun: ansys.api.speos.scene.v2.scene_pb2.ManualSun
            Wavelengths range protobuf object to modify.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_sun_manual method available in
        source classes.
        """

        def __init__(
            self,
            sun: scene_pb2.ManualSun,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                raise RuntimeError(
                    "BaseSourceAmbient.Manual class instantiated outside of class scope"
                )
            self._sun = sun

            if default_values:
                self.direction = [0, 0, 1]
                self.reverse_sun = False

        @property
        def direction(self) -> List[float]:
            """Get direction of the manual sun.

            Returns
            -------
            list of float
                list describing the direction of the manual sun.

            """
            return self._sun.sun_direction

        @direction.setter
        def direction(self, direction: List[float]) -> None:
            """Set direction of the manual sun.

                default value to be [0, 0, 1].

            Parameters
            ----------
            direction: List[float]
                direction of the sun.

            Returns
            -------
            BaseSourceAmbient.Manual

            """
            self._sun.sun_direction[:] = direction

        @property
        def reverse_sun(self) -> bool:
            """Get whether reverse direction of the manual sun.

            Returns
            -------
            bool
                True to reverse direction, False to not reverse direction

            """
            return self._sun.reverse_sun

        @reverse_sun.setter
        def reverse_sun(self, value: bool) -> None:
            """Reverse direction of the manual sun.

                default value to be False.

            Parameters
            ----------
            value: bool
                True to reverse direction, False to not reverse direction

            Returns
            -------
            None

            """
            self._sun.reverse_sun = value


class SourceAmbientNaturalLight(BaseSourceAmbient):
    """Natural light ambient source.

    By default, turbidity is set to be 3 with Sky.
    [0, 0, 1] is used as zenith direction, [0, 1, 0] as north direction.
    Sun type is set to be automatic type.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    default_values : bool
        Uses default values when True.
    """

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            source_instance=source_instance,
        )
        self._speos_client = self._project.client
        self._name = name
        self._type = None

        if default_values:
            # Default values
            self.zenith_direction = [0, 0, 1]
            self.north_direction = [0, 1, 0]
            self.reverse_north_direction = False
            self.reverse_zenith_direction = False
            self.turbidity = 3
            self.with_sky = True
            self.set_sun_automatic()

    @property
    def turbidity(self) -> float:
        """Get turbidity of the natural light source.

        Returns
        -------
        float
            value of Turbidity the measure of the fraction of scattering.

        """
        return self._source_template.ambient.natural_light.turbidity

    @turbidity.setter
    def turbidity(self, value: float) -> None:
        """Set turbidity of the natural light source.

            default value to be 3.

        Parameters
        ----------
        value: float
            set value of Turbidity the measure of the fraction of scattering.

        Returns
        -------
        None

        """
        if not 1.9 <= value <= 9.9:
            raise ValueError("Varies needs to be between 1.9 and 9.9")
        self._source_template.ambient.natural_light.turbidity = value

    @property
    def with_sky(self) -> bool:
        """Bool of whether activated using sky in the natural light source.

        Returns
        -------
        bool
            True as using sky, while False as using natural light without the sky.

        """
        return self._source_template.ambient.natural_light.with_sky

    @with_sky.setter
    def with_sky(self, value: bool) -> None:
        """Activate using sky in the natural light source.

            default value to be True.

        Parameters
        ----------
        value: bool
            True as using sky, while False as using natural light without the sky.

        Returns
        -------
        SourceAmbientNaturalLight

        """
        self._source_template.ambient.natural_light.with_sky = value

    @property
    def zenith_direction(self) -> List[float]:
        """Get zenith direction of the natural light source.

        Returns
        -------
        List[float]
            direction defines the zenith direction of the natural light.

        """
        return self._source_instance.ambient_properties.zenith_direction

    @zenith_direction.setter
    def zenith_direction(self, direction: Optional[List[float]] = None) -> None:
        """Set zenith direction of the natural light source.

            default value to be [0, 0, 1]

        Parameters
        ----------
        direction: Optional[List[float]]
            direction defines the zenith direction of the natural light.

        Returns
        -------
        None

        """
        self._source_instance.ambient_properties.zenith_direction[:] = direction

    @property
    def reverse_zenith_direction(self) -> bool:
        """
        Get whether reverse zenith direction of the natural light source.

        Returns
        -------
        bool
            True to reverse zenith direction, False otherwise.

        """
        return self._source_instance.ambient_properties.reverse_zenith

    @reverse_zenith_direction.setter
    def reverse_zenith_direction(self, value: bool) -> None:
        """Set reverse zenith direction of the natural light source.

            default value to be False.

        Parameters
        ----------
        value: bool
            True to reverse zenith direction, False otherwise.

        Returns
        -------
        None

        """
        self._source_instance.ambient_properties.reverse_zenith = value

    @property
    def north_direction(self) -> List[float]:
        """Get north direction of the natural light source.

        Returns
        -------
        List[float]
            direction defines the north direction of the natural light.

        """
        return self._source_instance.ambient_properties.natural_light_properties.north_direction

    @north_direction.setter
    def north_direction(self, direction: List[float]) -> None:
        """Set north direction of the natural light source.

            default value to be [0, 1, 0].

        Parameters
        ----------
        direction: List[float]
            direction defines the north direction of the natural light.

        Returns
        -------
        None

        """
        self._source_instance.ambient_properties.natural_light_properties.north_direction[:] = (
            direction
        )

    @property
    def reverse_north_direction(self) -> bool:
        """Get whether reverse north direction of the natural light source.

        Returns
        -------
        bool
            True as reverse north direction, False otherwise.

        """
        return self._source_instance.ambient_properties.natural_light_properties.reverse_north

    @reverse_north_direction.setter
    def reverse_north_direction(self, value: bool) -> None:
        """Set reverse north direction of the natural light source.

            default value to be False.

        Parameters
        ----------
        value: bool
            True to reverse north direction, False otherwise.

        Returns
        -------
        None

        """
        self._source_instance.ambient_properties.natural_light_properties.reverse_north = value

    def set_sun_automatic(self) -> BaseSourceAmbient.AutomaticSun:
        """Set natural light sun type as automatic.

        Returns
        -------
        BaseSourceAmbient.AutomaticSun

        """
        natural_light_properties = self._source_instance.ambient_properties.natural_light_properties
        if self._type is None and natural_light_properties.sun_axis_system.HasField(
            "automatic_sun"
        ):
            self._type = BaseSourceAmbient.AutomaticSun(
                natural_light_properties.sun_axis_system.automatic_sun,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._type, BaseSourceAmbient.AutomaticSun):
            # if the _type is not Colorimetric then we create a new type.
            self._type = BaseSourceAmbient.AutomaticSun(
                natural_light_properties.sun_axis_system.automatic_sun,
                stable_ctr=True,
            )
        elif self._type._sun is not natural_light_properties.sun_axis_system.automatic_sun:
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._sun = natural_light_properties.sun_axis_system.automatic_sun
        return self._type

    def set_sun_manual(self) -> BaseSourceAmbient.Manual:
        """Set natural light sun type as manual.

        Returns
        -------
        BaseSourceAmbient.Manual
        """
        natural_light_properties = self._source_instance.ambient_properties.natural_light_properties
        if self._type is None and natural_light_properties.sun_axis_system.HasField("manual_sun"):
            self._type = BaseSourceAmbient.Manual(
                natural_light_properties.sun_axis_system.manual_sun,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._type, BaseSourceAmbient.Manual):
            # if the _type is not Colorimetric then we create a new type.
            self._type = BaseSourceAmbient.Manual(
                natural_light_properties.sun_axis_system.manual_sun,
                stable_ctr=True,
            )
        elif self._type._sun is not natural_light_properties.sun_axis_system.manual_sun:
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._sun = natural_light_properties.sun_axis_system.manual_sun
        return self._type


class SourceAmbientEnvironment(BaseSourceAmbient):
    """Environment ambient source.

    By default [0, 0, 1] is used as zenith direction, [0, 1, 0] as north direction.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    default_values : bool
        Uses default values when True.
    """

    # source_type = "SourceAmbientEnvironment"
    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        source_instance: Optional[ProtoScene.SourceInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            source_instance=source_instance,
        )
        self._speos_client = self._project.client
        self._name = name
        self._type = None

        if default_values:
            # Default values
            self.zenith_direction = [0, 0, 1]
            self.north_direction = [0, 1, 0]
            self.reverse_north_direction = False
            self.reverse_zenith_direction = False
            self.luminance = 1000
            self.set_predefined_color_space()  # defaults to sRGB

    @property
    def zenith_direction(self) -> List[float]:
        """Zenith direction of the environment light source.

        This property get and set the zenith direction of the environment source.

        Parameters
        ----------
        direction: Optional[List[float]]
            direction defines the zenith direction of the environment light source.

        Returns
        -------
        List[float]
            direction defines the zenith direction of the environment light source.

        """
        return self._source_instance.ambient_properties.zenith_direction

    @zenith_direction.setter
    def zenith_direction(self, direction: Optional[List[float]]) -> None:
        self._source_instance.ambient_properties.zenith_direction[:] = direction

    @property
    def reverse_zenith_direction(self) -> bool:
        """Reverse zenith direction of the environment light source.

        This property get and set if reverse zenith direction is True.


        Parameters
        ----------
        value: bool
            True to reverse zenith direction, False otherwise.

        Returns
        -------
        bool
            True to reverse zenith direction, False otherwise.

        """
        return self._source_instance.ambient_properties.reverse_zenith

    @reverse_zenith_direction.setter
    def reverse_zenith_direction(self, value: bool) -> None:
        self._source_instance.ambient_properties.reverse_zenith = value

    @property
    def north_direction(self) -> List[float]:
        """North direction of the environment light source.

        This property get and set the north direction of the environment source.

        Parameters
        ----------
        direction: List[float]
            direction defines the north direction, default value to be [0, 1, 0].

        Returns
        -------
        List[float]
            direction defines the north direction of the environment source.

        """
        return self._source_instance.ambient_properties.environment_map_properties.north_direction

    @north_direction.setter
    def north_direction(self, direction: List[float]) -> None:
        self._source_instance.ambient_properties.environment_map_properties.north_direction[:] = (
            direction
        )

    @property
    def reverse_north_direction(self) -> bool:
        """Reverse north direction of the environment light source.

        This property get and set if reverse north direction is True.

        Parameters
        ----------
        value: bool
            True to reverse north direction, False otherwise.

        Returns
        -------
        bool
            True as reverse north direction, False otherwise.

        """
        return self._source_instance.ambient_properties.environment_map_properties.reverse_north

    @reverse_north_direction.setter
    def reverse_north_direction(self, value: bool) -> None:
        self._source_instance.ambient_properties.environment_map_properties.reverse_north = value

    @property
    def luminance(self) -> float:
        """Luminance of the environment light source.

        This property get and set the Luminance value of the source.

        Parameters
        ----------
        value: float
            set value of Luminance (cd/m^2).

        Returns
        -------
        float
            value of Luminance setting (cd/m^2).

        """
        return self._source_template.ambient.environment_map.luminance

    @luminance.setter
    def luminance(self, value: float) -> None:
        self._source_template.ambient.environment_map.luminance = value

    @property
    def image_file_uri(self) -> str:
        """Location of the environment image file.

        This property gets or sets the environment image file used by the
        ambient environment source.

        Parameters
        ----------
         uri : Union[str, Path]
            format file uri (hdr, exr, png, bmp, jpg, tiff, rgb).

        Returns
        -------
        uri : str
            format file uri (hdr, exr, png, bmp, jpg, tiff, rgb).
        """
        return self._source_template.ambient.environment_map.image_uri

    @image_file_uri.setter
    def image_file_uri(self, uri: Union[str, Path]) -> None:
        self._source_template.ambient.environment_map.image_uri = str(uri)

    @property
    def color_space(
        self,
    ) -> Union[
        None,
        SourceAmbientEnvironment.PredefinedColorSpace,
        SourceAmbientEnvironment.UserDefinedColorSpace,
    ]:
        """Property containing all options in regard to the color space properties.

        Returns
        -------
        Union[
            None,
            ansys.speos.core.source.SourceAmbientEnvironment.PredefinedColorSpace,
            ansys.speos.core.source.SourceAmbientEnvironment.UserDefinedColorSpace
            ]
            Instance of Predefined Color Space class
        """
        return self._type

    def set_userdefined_color_space(self) -> SourceAmbientEnvironment.UserDefinedColorSpace:
        """Set the color space to user-defined.

        Returns
        -------
        SourceAmbientEnvironment.UserDefinedColorSpace
            Settings for user defined color space.

        """
        if self._type is None and self._source_template.ambient.environment_map.HasField(
            "user_defined_rgb_space"
        ):
            self._type = SourceAmbientEnvironment.UserDefinedColorSpace(
                project=self._project,
                userdefined_color_space=self._source_template.ambient.environment_map.user_defined_rgb_space,
                default_values=False,
                stable_ctr=True,
            )
        if not isinstance(self._type, SourceAmbientEnvironment.UserDefinedColorSpace):
            # if the _type is not UserDefinedColorSpace then we create a new type.
            self._type = SourceAmbientEnvironment.UserDefinedColorSpace(
                project=self._project,
                userdefined_color_space=self._source_template.ambient.environment_map.user_defined_rgb_space,
                stable_ctr=True,
            )
        elif (
            self._type._userdefined_color_space
            is not self._source_template.ambient.environment_map.user_defined_rgb_space
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._userdefined_color_space = (
                self._source_template.ambient.environment_map.user_defined_rgb_space
            )
        return self._type

    def set_predefined_color_space(self) -> SourceAmbientEnvironment.PredefinedColorSpace:
        """Set the color space to use one of the presets.

        Returns
        -------
        ansys.speos.core.source.SourceAmbientEnvironment.PredefinedColorSpace
            Environment source color space for sRGB or AdobeRGB
        """
        if self._type is None and self._source_template.ambient.environment_map.HasField(
            "predefined_color_space"
        ):
            self._type = SourceAmbientEnvironment.PredefinedColorSpace(
                predefined_color_space=self._source_template.ambient.environment_map.predefined_color_space,
                default_values=False,
                stable_ctr=True,
            )
        if not isinstance(self._type, SourceAmbientEnvironment.PredefinedColorSpace):
            # if the _type is not PredefinedColorSpace then we create a new type.
            self._type = SourceAmbientEnvironment.PredefinedColorSpace(
                predefined_color_space=self._source_template.ambient.environment_map.predefined_color_space,
                stable_ctr=True,
            )
        elif (
            self._type._predefined_color_space
            is not self._source_template.ambient.environment_map.predefined_color_space
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._predefined_color_space = (
                self._source_template.ambient.environment_map.predefined_color_space
            )
        return self._type

    def commit(self) -> SourceAmbientEnvironment:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.source.SourceAmbientEnvironment
            Ambient environment Source feature.
        """
        if isinstance(self._type, BaseSource.UserDefinedColorSpace):
            self._type._red_spectrum._commit()
            self._type._green_spectrum._commit()
            self._type._blue_spectrum._commit()
        super().commit()

    def reset(self) -> SourceAmbientEnvironment:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.source.SourceAmbientEnvironment
            Ambient environment Source feature.
        """
        if isinstance(self._type, BaseSource.UserDefinedColorSpace):
            self._type._red_spectrum._reset()
            self._type._green_spectrum._reset()
            self._type._blue_spectrum._reset()
        super().reset()

    def delete(self) -> SourceAmbientEnvironment:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.source.SourceAmbientEnvironment
            Ambient environment Source feature.
        """
        if isinstance(self._type, BaseSource.UserDefinedColorSpace):
            self._type._red_spectrum._delete()
            self._type._green_spectrum._delete()
            self._type._blue_spectrum._delete()
        super().delete()



# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Open one of the possible results generated out of the simulation."""

import os
from pathlib import Path
import tempfile
from typing import List, Union

import ansys.api.speos.file.v1.file_transfer as file_transfer_helper__v1
import ansys.api.speos.file.v1.file_transfer_pb2_grpc as file_transfer__v1__pb2_grpc
from ansys.api.speos.part.v1 import face_pb2

if os.name == "nt":
    from comtypes.client import CreateObject


import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import numpy

from ansys.speos.core.simulation import (
    SimulationDirect,
    SimulationInteractive,
    SimulationInverse,
)


class _Speos3dData:
    def __init__(
        self,
        x,
        y,
        z,
        illuminance=0.0,
        irradiance=0.0,
        reflection=0.0,
        transmission=0.0,
        absorption=0.0,
    ):
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)
        self.illuminance = float(illuminance)
        self.irradiance = float(irradiance)
        self.reflection = float(reflection)
        self.transmission = float(transmission)
        self.absorption = float(absorption)


def _find_correct_result(
    simulation_feature: Union[SimulationDirect, SimulationInverse, SimulationInteractive],
    result_name: str,
    download_if_distant: bool = True,
) -> str:
    if len(simulation_feature.result_list) == 0:
        raise ValueError("Please compute the simulation feature to generate results.")

    file_path = ""

    for res in simulation_feature.result_list:
        if res.HasField("path"):
            if res.path.endswith(result_name):
                file_path = res.path
                break
        elif res.HasField("upload_response"):
            if res.upload_response.info.file_name == result_name:
                if download_if_distant:
                    file_transfer_helper__v1.download_file(
                        file_transfer_service_stub=file_transfer__v1__pb2_grpc.FileTransferServiceStub(
                            simulation_feature._project.client.channel
                        ),
                        file_uri=res.upload_response.info.uri,
                        download_location=tempfile.gettempdir(),
                    )
                    file_path = str(
                        Path(tempfile.gettempdir()) / res.upload_response.info.file_name
                    )
                else:
                    file_path = res.upload_response.info.uri
                break
    return file_path


def _display_image(img: numpy.ndarray):
    if img is not None:
        plt.imshow(img)
        plt.axis("off")  # turns off axes
        plt.axis("tight")  # gets rid of white border
        plt.axis("image")  # square up the image instead of filling the "figure" space
        plt.show()


if os.name == "nt":

    def open_result_image(
        simulation_feature: Union[SimulationDirect, SimulationInverse, SimulationInteractive],
        result_name: str,
    ) -> None:
        """Retrieve an image from a specific simulation result.

        Parameters
        ----------
        simulation_feature : ansys.speos.core.simulation.Simulation
            The simulation feature.
        result_name : str
            The result name to open as an image.
        """
        file_path = _find_correct_result(simulation_feature, result_name)
        if file_path == "":
            raise ValueError(
                "No result corresponding to "
                + result_name
                + " is found in "
                + simulation_feature._name
            )

        if file_path.endswith("xmp") or file_path.endswith("XMP"):
            dpf_instance = CreateObject("XMPViewer.Application")
            dpf_instance.OpenFile(file_path)
            res = dpf_instance.ExportXMPImage(file_path + ".png", 1)
            if res:
                _display_image(mpimg.imread(file_path + ".png"))
        elif file_path.endswith("png") or file_path.endswith("PNG"):
            _display_image(mpimg.imread(file_path))

    def open_result_in_viewer(
        simulation_feature: Union[SimulationDirect, SimulationInverse],
        result_name: str,
    ) -> None:
        """Open a specific simulation result in the suitable viewer.

        Parameters
        ----------
        simulation_feature : ansys.speos.core.simulation.Simulation
            The simulation feature.
        result_name : str
            The result name to open in a viewer.
        """
        file_path = _find_correct_result(simulation_feature, result_name)

        if file_path.endswith("xmp") or file_path.endswith("XMP"):
            dpf_instance = CreateObject("XMPViewer.Application")
            dpf_instance.OpenFile(file_path)
            dpf_instance.Show(1)
        elif file_path.endswith("hdr") or file_path.endswith("HDR"):
            dpf_instance = CreateObject("HDRIViewer.Application")
            dpf_instance.OpenFile(file_path)
            dpf_instance.Show(1)

    def export_xmp_vtp(
        simulation_feature: Union[SimulationDirect, SimulationInverse],
        result_name: Union[str, Path],
    ) -> Path:
        """Export an XMP result into vtp file.

        Parameters
        ----------
        simulation_feature : ansys.speos.core.simulation.Simulation
            The simulation feature.
        result_name: Union[str, Path]
            file path of an XMP result.

        Returns
        -------
        Path
            file path of exported vtp file.

        """
        import pyvista as pv

        result_name = Path(result_name)
        if not str(result_name).lower().endswith(".xmp"):
            result_name = result_name.with_name(result_name.name + ".xmp")
        file_path = _find_correct_result(simulation_feature, str(result_name))

        if file_path == "":
            raise ValueError(
                "No result corresponding to "
                + str(result_name)
                + " is found in "
                + simulation_feature._name
            )

        file_path = Path(file_path)
        dpf_instance = CreateObject("XMPViewer.Application")
        dpf_instance.OpenFile(str(file_path))
        dimension_x = dpf_instance.XWidth
        dimension_y = dpf_instance.YHeight
        resolution_x = dpf_instance.XNb
        resolution_y = dpf_instance.YNb
        tmp_txt = file_path.with_suffix(".txt")
        dpf_instance.ExportTXT(str(tmp_txt))

        file = tmp_txt.open("r")
        content = file.readlines()
        file.close()
        skip_lines = 9 if "SeparatedByLayer" in content[7] else 8
        xmp_data = []
        if dpf_instance.Maptype == 2 and len(content[6].strip().split()) == 3:
            # spectral data within number of data tables
            spectral_tables = int(content[6].strip().split()[2])
            xmp_data = [
                [0 for _ in range(len(content[skip_lines].strip().split()))]
                for _ in range(resolution_y)
            ]
            for _ in range(spectral_tables):
                for i in range(resolution_y):
                    row = list(map(float, content[skip_lines].strip().split()))
                    for j in range(resolution_x):
                        xmp_data[i][j] += row[j]
                    skip_lines += 1
                # Skip one line between tables
                skip_lines += 1
        else:
            # not spectral data
            for line in content[skip_lines : skip_lines + resolution_y]:
                line_content = line.strip().split()
                xmp_data.append(list(map(float, line_content)))

        # Create VTK ImageData structure
        step_x = float(dimension_x) / resolution_x
        step_y = float(dimension_y) / resolution_y
        origin_x = -(resolution_x * step_x) / 2
        origin_y = -(resolution_y * step_y) / 2
        grid = pv.ImageData(
            dimensions=(resolution_x, resolution_y, 1),
            spacing=(step_x, step_y, 1),
            origin=(origin_x, origin_y, 0),
        )
        xmp_data = numpy.array(xmp_data)
        if xmp_data.shape[1] == resolution_x:
            if dpf_instance.UnitType == 0:
                grid["Radiometric"] = numpy.ravel(xmp_data)
            if dpf_instance.UnitType == 1:
                grid["Photometric"] = numpy.ravel(xmp_data)
        else:
            grid["X"] = numpy.ravel(xmp_data[:, 0::4])
            grid["Photometric"] = numpy.ravel(xmp_data[:, 1::4])
            grid["Radiometric"] = numpy.ravel(xmp_data[:, 2::4])
            grid["Z"] = numpy.ravel(xmp_data[:, 3::4])
        vtp_meshes = grid.extract_surface()
        # Export file to VTP
        vtp_meshes.save(str(file_path.with_suffix(".vtp")))
        return file_path.with_suffix(".vtp")

    def export_xm3_vtp(
        simulation_feature: Union[SimulationDirect, SimulationInverse],
        geo_faces: List[face_pb2.Face],
        result_name: Union[str, Path],
    ) -> Path:
        """Export an XMP result into vtp file.

        Parameters
        ----------
        simulation_feature : ansys.speos.core.simulation.Simulation
            The simulation feature.
        geo_faces: List[face_pb2.Face]
            list of face geometries.
        result_name: Union[str, Path]
            file path of an XMP result.

        Returns
        -------
        Path
            file path of exported vtp file.

        """
        import pyvista as pv

        result_name = Path(result_name)
        if not str(result_name).lower().endswith(".xm3"):
            result_name = result_name.with_name(result_name.name + ".xm3")
        file_path = _find_correct_result(simulation_feature, str(result_name))

        if file_path == "":
            raise ValueError(
                "No result corresponding to "
                + str(result_name)
                + " is found in "
                + simulation_feature._name
            )

        file_path = Path(file_path)
        dpf_instance = CreateObject("Xm3Viewer.Application")
        dpf_instance.OpenFile(str(file_path))
        tmp_txt = file_path.with_suffix(".txt")
        dpf_instance.Export(str(tmp_txt))

        file = tmp_txt.open("r")
        xm3_data = []
        content = file.readlines()
        header = content[0].strip().split("\t")
        illuminance_indices = [
            i for i, header_item in enumerate(header) if header_item == "Illuminance"
        ]
        irradiance_indices = [
            i for i, header_item in enumerate(header) if header_item == "Irradiance"
        ]
        reflection_indices = [
            i for i, header_item in enumerate(header) if "Reflection" in header_item
        ]
        transmission_indices = [
            i for i, header_item in enumerate(header) if "Transmission" in header_item
        ]
        absorption_indices = [
            i for i, header_item in enumerate(header) if "Absorption" in header_item
        ]

        skip_line = 1
        try:
            float(float(content[1].strip().split()[0]))
            skip_line = 1  # only single layer
        except ValueError:
            skip_line = 2  # separated layer
        for line in content[skip_line:]:
            line_content = line.strip().split("\t")
            xm3_data.append(
                _Speos3dData(
                    x=float(line_content[0]),
                    y=float(line_content[1]),
                    z=float(line_content[2]),
                    illuminance=sum(
                        [
                            float(item)
                            for i, item in enumerate(line_content)
                            if i in illuminance_indices
                        ],
                        0.0,
                    ),
                    irradiance=sum(
                        [
                            float(item)
                            for i, item in enumerate(line_content)
                            if i in irradiance_indices
                        ],
                        0.0,
                    ),
                    reflection=sum(
                        [
                            float(item)
                            for i, item in enumerate(line_content)
                            if i in reflection_indices
                        ],
                        0.0,
                    ),
                    transmission=sum(
                        [
                            float(item)
                            for i, item in enumerate(line_content)
                            if i in transmission_indices
                        ],
                        0.0,
                    ),
                    absorption=sum(
                        [
                            float(item)
                            for i, item in enumerate(line_content)
                            if i in absorption_indices
                        ],
                        0.0,
                    ),
                )
            )

        vtp_meshes = None
        for geo in geo_faces:
            vertices = numpy.array(geo.vertices).reshape(-1, 3)
            facets = numpy.array(geo.facets).reshape(-1, 3)
            temp = numpy.full(facets.shape[0], 3)
            temp = numpy.vstack(temp)
            facets = numpy.hstack((temp, facets))
            if vtp_meshes is None:
                vtp_meshes = pv.PolyData(vertices, facets)
            else:
                vtp_meshes = vtp_meshes.append_polydata(pv.PolyData(vertices, facets))

        vtp_meshes["Illuminance [lx]"] = [item.illuminance for item in xm3_data]
        vtp_meshes["Irradiance [W/m2]"] = [item.irradiance for item in xm3_data]
        vtp_meshes["Reflection"] = [item.reflection for item in xm3_data]
        vtp_meshes["Transmission"] = [item.transmission for item in xm3_data]
        vtp_meshes["Absorption"] = [item.absorption for item in xm3_data]
        # vtp_meshes = vtp_meshes.point_data_to_cell_data()
        vtp_meshes.save(str(file_path.with_suffix(".vtp")))
        return file_path.with_suffix(".vtp")
